import numpy as np
import pandas as pd
from anndata import AnnData
from scipy import sparse
import networkx as nx
import scanpy as sc
from typing import Optional, Union
from pathlib import Path
import re

import matplotlib.pyplot as plt
from matplotlib.pyplot import gcf
import matplotlib.ticker as ticker
import seaborn as sns

from .resources import TFs_human, TFs_mouse


def data_preparation(input_expData: Union[str, sc.AnnData, pd.DataFrame],
                     input_priorNet: Union[str, pd.DataFrame],
                     genes_DE: Optional[Union[str, pd.DataFrame, pd.Series]] = None,
                     additional_edges_pct: float = 0.01):
    print('[0] - Data loading and preprocessing...')

    ## [1] Single-cell RNA-seq data
    lineages = None
    if isinstance(input_expData, str):
        p = Path(input_expData)
        if p.suffix == '.csv':
            adata = sc.read_csv(input_expData, first_column_names=True)
        else:  # h5ad
            adata = sc.read_h5ad(input_expData)
    elif isinstance(input_expData, sc.AnnData):
        adata = input_expData
        lineages = adata.uns.get('lineages')
    elif isinstance(input_expData, pd.DataFrame):
        adata = sc.AnnData(X=input_expData)
    else:
        raise Exception("Invalid input! The input format must be '.csv' file or '.h5ad' "
                        "formatted file, or an 'AnnData' object!", input_expData)

    possible_species = 'mouse' if bool(re.search('[a-z]', adata.var_names[0])) else 'human'

    # Gene symbols are uniformly handled in uppercase
    adata.var_names = adata.var_names.str.upper()

    ## [2] Prior network data
    if isinstance(input_priorNet, str):
        netData = pd.read_csv(input_priorNet, index_col=None, header=0)
    elif isinstance(input_priorNet, pd.DataFrame):
        netData = input_priorNet.copy()
    else:
        raise Exception("Invalid input!", input_priorNet)
    # make sure the genes of prior network are in the input scRNA-seq data
    netData['from'] = netData['from'].str.upper()
    netData['to'] = netData['to'].str.upper()
    netData = netData.loc[netData['from'].isin(adata.var_names.values)
                          & netData['to'].isin(adata.var_names.values), :]
    netData = netData.drop_duplicates(subset=['from', 'to'], keep='first', inplace=False)

    # Transfer into networkx object
    priori_network = nx.from_pandas_edgelist(netData, source='from', target='to', create_using=nx.DiGraph)
    priori_network_nodes = np.array(priori_network.nodes())

    # in_degree, out_degree (centrality)
    in_degree = pd.DataFrame.from_dict(nx.in_degree_centrality(priori_network),
                                       orient='index', columns=['in_degree'])
    out_degree = pd.DataFrame.from_dict(nx.out_degree_centrality(priori_network),
                                        orient='index', columns=['out_degree'])
    centrality = pd.concat([in_degree, out_degree], axis=1)
    centrality = centrality.loc[priori_network_nodes, :]

    ## [3] A mapper for node index and gene name
    idx_GeneName_map = pd.DataFrame({'idx': range(len(priori_network_nodes)),
                                     'geneName': priori_network_nodes},
                                    index=priori_network_nodes)

    edgelist = pd.DataFrame({'from': idx_GeneName_map.loc[netData['from'].tolist(), 'idx'].tolist(),
                             'to': idx_GeneName_map.loc[netData['to'].tolist(), 'idx'].tolist()})

    ## [4] add TF information
    is_TF = np.ones(len(priori_network_nodes), dtype=int)
    if possible_species == 'human':
        TFs_df = TFs_human
    else:
        TFs_df = TFs_mouse
    TF_list = TFs_df.iloc[:, 0].str.upper()
    is_TF[~np.isin(priori_network_nodes, TF_list)] = 0

    # Only keep the genes that exist in both single cell data and the prior gene interaction network
    adata = adata[:, priori_network_nodes]
    if lineages is None:
        cells_in_lineage_dict = {'all': adata.obs_names}  # all cells are regarded as in on lineage if lineages is None
    else:
        cells_in_lineage_dict = {l: adata.obs_names[adata.obs[l].notna()] for l in lineages}

    print(f"Consider the input data with {len(cells_in_lineage_dict)} lineages:")
    adata_lineages = dict()
    for l, c in cells_in_lineage_dict.items():
        print(f"  Lineage - {l}:")

        adata_l = sc.AnnData(X=adata[c, :].to_df())
        adata_l.var['is_TF'] = is_TF
        adata_l.varm['centrality_prior_net'] = centrality
        adata_l.varm['idx_GeneName_map'] = idx_GeneName_map
        adata_l.uns['name'] = l

        ## [5] Additional edges with high spearman correlation
        if isinstance(adata_l.X, sparse.csr_matrix):
            gene_exp = pd.DataFrame(adata_l.X.A.T, index=priori_network_nodes)
        else:
            gene_exp = pd.DataFrame(adata_l.X.T, index=priori_network_nodes)

        ori_edgeNum = len(edgelist)
        edges_corr = np.absolute(np.array(gene_exp.T.corr('spearman')))
        np.fill_diagonal(edges_corr, 0.0)
        x, y = np.where(edges_corr > 0.6)
        addi_top_edges = pd.DataFrame({'from': x, 'to': y, 'weight': edges_corr[x, y]})
        addi_top_k = int(gene_exp.shape[0] * (gene_exp.shape[0] - 1) * additional_edges_pct)
        if len(addi_top_edges) > addi_top_k:
            addi_top_edges = addi_top_edges.sort_values(by=['weight'], ascending=False)
            addi_top_edges = addi_top_edges.iloc[0:addi_top_k, 0:2]
        edgelist = pd.concat([edgelist, addi_top_edges.iloc[:, 0:2]], ignore_index=True)
        edgelist = edgelist.drop_duplicates(subset=['from', 'to'], keep='first', inplace=False)
        print('    {} extra edges (Spearman correlation > 0.6) are added into the prior gene interaction network.\n'
              '    Total number of edges: {}.'.format((len(edgelist) - ori_edgeNum), len(edgelist)))

        adata_l.uns['edgelist'] = edgelist

        ## [6] Differential expression scores
        logFC = adata.var.get(l + '_logFC')
        if (genes_DE is None) and (logFC is None):
            pass
        else:
            if genes_DE is not None:
                if isinstance(genes_DE, str):
                    genes_DE = pd.read_csv(genes_DE, index_col=0, header=0)
            else:  # logFC is not None
                genes_DE = logFC

            genes_DE = pd.DataFrame(genes_DE).iloc[:, 0]
            genes_DE.index = genes_DE.index.str.upper()
            genes_DE = genes_DE[genes_DE.index.isin(priori_network_nodes)].abs().dropna()
            node_score_auxiliary = pd.Series(np.zeros(len(priori_network_nodes)), index=priori_network_nodes)
            node_score_auxiliary[genes_DE.index] = genes_DE.values
            node_score_auxiliary = np.array(node_score_auxiliary)
            adata_l.var['node_score_auxiliary'] = node_score_auxiliary
            genes_DE = None

        adata_lineages[l] = adata_l
        print(f"    n_genes × n_cells = {adata_l.n_vars} × {adata_l.n_obs}")

    return adata_lineages


def cluster_cell_by_RGM(auc_mtx, true_cell_label, method='ward', k=None):
    """
    Cluster cells based on RGM activity matrix by using hierarchical clustering.
    """
    assert method in {'ward', 'complete', 'average', 'single'}
    from sklearn.cluster import AgglomerativeClustering
    from sklearn.metrics import silhouette_score, normalized_mutual_info_score, adjusted_rand_score

    auc_mtx_Z = pd.DataFrame(index=auc_mtx.index, columns=list(auc_mtx.columns))
    for row in list(auc_mtx.index):
        auc_mtx_Z.loc[row, :] = (auc_mtx.loc[row, :] - auc_mtx.loc[row, :].mean()) / auc_mtx.loc[row, :].std(ddof=0)

    if k is not None:
        ac = AgglomerativeClustering(n_clusters=k, affinity='euclidean', linkage=method).fit(auc_mtx_Z)
        predicted_cell_label = ac.labels_
        NMIs = normalized_mutual_info_score(true_cell_label, predicted_cell_label)
        ARIs = adjusted_rand_score(true_cell_label, predicted_cell_label)
        Silhouettes = silhouette_score(auc_mtx, predicted_cell_label, metric='euclidean')
        N_clus = k
    else:
        NMIs, ARIs, N_clus, Silhouettes = [], [], [], []
        max_cluster_num = len(set(true_cell_label)) * 2
        for i in range(2, max_cluster_num):
            ac = AgglomerativeClustering(n_clusters=i, affinity='euclidean', linkage=method).fit(auc_mtx_Z)
            out = ac.labels_
            NMIs.append(normalized_mutual_info_score(true_cell_label, out))
            ARIs.append(adjusted_rand_score(true_cell_label, out))
            N_clus.append(i)
            Silhouettes.append(silhouette_score(auc_mtx, out, metric='euclidean'))

    return {'NMIs': NMIs, 'ARIs': ARIs, 'Silhouettes': Silhouettes, 'num_clusters': N_clus}


def network_topological_properties(prior_network: pd.DataFrame):
    import networkx as nx
    from sklearn.linear_model import LinearRegression as lr

    prior_nx = nx.from_pandas_edgelist(prior_network, source='from', target='to', edge_attr=None,
                                       create_using=nx.DiGraph)

    # Number of genes
    N_genes = prior_nx.number_of_nodes()

    # Number of edges
    N_edges = len(prior_network[['from', 'to']].drop_duplicates())

    # Number of source genes
    N_source_genes = len(prior_network['from'].unique())

    # Number of target genes
    N_target_genes = len(prior_network['to'].unique())

    # Density
    Density = nx.density(prior_nx)

    # Average degree / in_degree / out_degree
    degree = sum(dict(prior_nx.degree()).values()) / N_genes

    # Clustering coefficient (time-consuming)
    clustering_coefficient = nx.average_clustering(prior_nx)

    # Slope of degree distribution
    degree_sequence = pd.DataFrame(np.array(prior_nx.degree))
    degree_sequence.columns = ["ind", "degree"]
    degree_sequence = degree_sequence.set_index("ind")
    dist = degree_sequence.degree.value_counts() / degree_sequence.degree.value_counts().sum()
    dist.index = dist.index.astype(int)

    x = np.log(dist.index.values).reshape([-1, 1])
    y = np.log(dist.values).reshape([-1, 1])

    model = lr()
    model.fit(x, y)

    slope = model.coef_[0][0]

    print(  # f"Dataset-species: {dataset}-{species}\n"
        f"Number of genes: {N_genes}\nNumber of edges: {N_edges}\n"
        f"Number of source genes: {N_source_genes}\nNumber of target genes: {N_target_genes}\n"
        f"Network density: {Density}\nAverage degree: {degree:.4f}\n"
        f"Average clustering coefficient: {clustering_coefficient:.4f}\n"
        f"Slope of degree distribution: {slope:.4f}")


def prepare_data_for_R(adata: sc.AnnData,
                       temp_R_dir: str,
                       reducedDim: Optional[str] = None,
                       cluster_label: Optional[str] = None):
    """
    Process the AnnData object and save the necessary data to files.
    These data files are prepared for running the `slingshot_MAST_script.R` or `MAST_script.R` scripts.
    """
    if 'log_transformed' not in adata.layers:
        raise ValueError(
            f'Did not find `log_transformed` in adata.layers.'
        )

    if isinstance(adata.layers['log_transformed'], sparse.csr_matrix):
        exp_normalized = adata.layers['log_transformed'].A
    else:
        exp_normalized = adata.layers['log_transformed']

    # The normalized and log transformed data is used for MAST
    normalized_counts = pd.DataFrame(exp_normalized,
                                     index=adata.obs_names,
                                     columns=adata.var_names)
    normalized_counts.to_csv(temp_R_dir + '/exp_normalized.csv', sep=',')

    # The reduced dimension data is used for Slingshot
    if reducedDim is not None:
        reducedDim_data = pd.DataFrame(adata.obsm[reducedDim], dtype='float32', index=None)
        reducedDim_data.to_csv(temp_R_dir + '/data_reducedDim.csv', index=None)
    else:
        if 'lineages' not in adata.uns:
            raise ValueError(
                f'Did not find `lineages` in adata.uns.'
            )
        else:
            pseudotime_all = pd.DataFrame(index=adata.obs_names)
            for li in adata.uns['lineages']:
                pseudotime_all[li] = adata.obs[li]
            pseudotime_all.to_csv(temp_R_dir + '/pseudotime_lineages.csv', index=True)

    # Cluster Labels (Leiden)
    if cluster_label is not None:
        cluster_labels = pd.DataFrame(adata.obs[cluster_label])
        cluster_labels.to_csv(temp_R_dir + '/clusters.csv')


def process_Slingshot_MAST_R(temp_R_dir: str,
                             split_num: int = 4,
                             start_cluster: int = 0,
                             end_cluster: Optional[list] = None):
    """
    Run the `slingshot_MAST_script.R` to get pseudotim and differential expression information for each lineage.
    """
    import subprocess
    import importlib.resources as res

    R_script_path = 'slingshot_MAST_script.R'
    with res.path('cefcon', R_script_path) as datafile:
        R_script_path = datafile

    path = Path(temp_R_dir)
    path.mkdir(exist_ok=path.exists(), parents=True)

    args = f'Rscript {R_script_path} {temp_R_dir} {split_num} {start_cluster}'
    if end_cluster is not None:
        args += f' {end_cluster}'
    print('Running Slingshot and MAST using: \'{}\'\n'.format(args))
    print('It will take a few minutes ...')
    with subprocess.Popen(args,
                          stdout=None, stderr=subprocess.PIPE,
                          shell=True) as p:
        out, err = p.communicate()
        if p.returncode == 0:
            print(f'Done. The results are saved in \'{temp_R_dir}\'.')
            print(f'      Trajectory (pseudotime) information: \'pseudotime_lineages.csv\'.')
            lineages = pd.read_csv(temp_R_dir + '/pseudotime_lineages.csv', index_col=0)
            lineages.columns
            print('      Differential expression information: ', end='')
            for l in lineages.columns:
                print(f'\'DEgenes_MAST_sp{split_num}_{l}.csv\' ', end='')
        else:
            print(f'Something error: returncode={p.returncode}.')


def process_MAST_R(temp_R_dir: str, split_num: int = 4):
    """
    Run the `MAST_script.R` to get differential expression information for each lineage.
    """
    import subprocess
    import importlib.resources as res

    R_script_path = 'MAST_script.R'
    with res.path('cefcon', R_script_path) as datafile:
        R_script_path = datafile

    path = Path(temp_R_dir)
    path.mkdir(exist_ok=path.exists(), parents=True)

    args = f'Rscript {R_script_path} {temp_R_dir} {split_num}'
    print('Running MAST using: \'{}\'\n'.format(args))
    print('It will take a few minutes ...')
    with subprocess.Popen(args,
                          stdout=None, stderr=subprocess.PIPE,
                          shell=True) as p:
        out, err = p.communicate()
        if p.returncode == 0:
            print(f'Done. The results are saved in \'{temp_R_dir}\'.')
            print(f'      Differential expression information: \'DEgenes_MAST_sp{split_num}_<x>.csv\'')
        else:
            print(f'Something error: returncode={p.returncode}.')


def plot_controllability_metrics(cefcon_results: Union[dict, list], return_value: bool = False):
    """

    """
    # calculate metrics
    con_df = pd.DataFrame(columns=['MDS_controllability_score', 'MFVS_controllability_score',
                                   'Jaccard_index', 'Driver_regulators_coverage', 'Lineage'])
    for k in cefcon_results:
        if isinstance(k, str):
            result = cefcon_results[k]
        else:
            result = k
        drivers_df = result.driver_regulator

        MFVS_driver_set = set(drivers_df.loc[drivers_df['is_MFVS_driver']].index)
        MDS_driver_set = set(drivers_df.loc[drivers_df['is_MDS_driver']].index)
        driver_regulators = set(drivers_df.loc[drivers_df['is_driver_regulator']].index)
        top_ranked_genes = driver_regulators.union(
            (set(drivers_df.index) - MFVS_driver_set.union(MDS_driver_set)))
        N_genes = result.n_genes

        # MDS controllability score
        MDS_con = 1 - len(MDS_driver_set) / N_genes
        # MFVS controllability score
        MFVS_con = 1 - len(MFVS_driver_set) / N_genes
        # Jaccard index
        Jaccard_con = len(MDS_driver_set.intersection(MFVS_driver_set)) / len(MDS_driver_set.union(MFVS_driver_set))
        # driver regulators coverage
        Critical_con = len(driver_regulators) / len(MDS_driver_set.union(MFVS_driver_set))

        con_df.loc[len(con_df)] = {'MDS_controllability_score': MDS_con,
                                   'MFVS_controllability_score': MFVS_con,
                                   'Jaccard_index': Jaccard_con,
                                   'Driver_regulators_coverage': Critical_con,
                                   'Lineage': result.name}

    con_df = pd.melt(con_df, id_vars=['Lineage'])

    # plot
    fig = plt.figure(figsize=(9, 2))
    sns.set_theme(style="ticks", font_scale=1.0)
    surrent_palette = sns.color_palette("Set1")

    # Controallability score
    con_df1 = con_df.loc[con_df['variable'].isin(['MDS_controllability_score', 'MFVS_controllability_score'])]
    fig.add_subplot(1, 3, 1)
    ax1 = sns.barplot(x="variable", y="value", hue="Lineage", palette=surrent_palette,
                      data=con_df1)
    ax1.set_xlabel('')
    ax1.set_xticklabels(['MDS', 'MFVS'], rotation=45, ha="right")
    ax1.set_ylabel('Controllability Score')
    ax1.set_ylim(con_df1['value'].min().round(1)-0.1, 1.0)
    sns.despine()
    ax1.get_legend().remove()

    # Jaccard index
    con_df2 = con_df.loc[con_df['variable'].isin(['Jaccard_index'])]
    fig.add_subplot(1, 3, 2)
    ax2 = sns.barplot(x="variable", y="value", hue="Lineage", palette=surrent_palette,
                      data=con_df2)
    ax2.set_xlabel('')
    ax2.set_xticklabels('')
    ax2.set_ylabel(r'Jaccard Index\nbetween MFVS & MDS')
    sns.despine()
    ax2.get_legend().remove()

    # Driver regulators coverage
    con_df3 = con_df.loc[con_df['variable'].isin(['Driver_regulators_coverage'])]
    fig.add_subplot(1, 3, 3)
    ax3 = sns.barplot(x="variable", y="value", hue="Lineage", palette=surrent_palette,
                      data=con_df3)
    ax3.set_xlabel('')
    ax3.set_xticklabels('')
    ax3.set_ylabel('Driver Regulators Coverage')
    sns.despine()
    plt.legend(loc='upper left', bbox_to_anchor=(1, 1), fontsize='small')

    plt.subplots_adjust(wspace=0.45)

    if return_value:
        return con_df



# -*- coding: utf-8 -*-
import torch
import numpy as np

def sparse_mx_to_torch_edge_list(sparse_mx):
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    edge_list = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    return edge_list

def corruption(x, edge_index):
    return x[torch.randperm(x.size(0))], edge_index


import numpy as np
import pandas as pd
from scipy.sparse import issparse
from matplotlib.colors import cnames, is_color_like, ListedColormap, to_rgb
from matplotlib import patheffects, rcParams
import scanpy as sc
import anndata
from typing import Union, Sequence

#from scvelo import logging as logg



def plot_heatmap(
    adata:anndata.AnnData,
    var_names: Union[str, Sequence[str]],
    sortby:str="latent_time",
    layer:str="Ms",
    color_map:str="RdBu_r",
    col_color=None,
    palette:str="viridis",
    n_convolve:int=30,
    standard_scale:int=0,
    sort:bool=True,
    colorbar=None,
    col_cluster:bool=False,
    row_cluster:bool=False,
    context=None,
    font_scale=None,
    figsize:tuple=(8, 4),
    show=None,
    save=None,
    **kwargs,
):
    r"""Plot time series for genes as heatmap.

    Arguments:
        adata: Annotated data matrix
        var_names: Names of variables to use for the plot
        sortby: Observation key to extract time data from ('latent_time')
        layer: Layer key to extract count data from ('Ms')
        color_map: String denoting matplotlib color map ('RdBu_r')
        col_color: String denoting matplotlib color map to use along columns (None)
        palette: Colors to use for plotting groups ('viridis')
        n_convolve: If int is given, data is smoothed by convolution (30)
        standard_scale: Either 0 (rows) or 1 (columns) for standardization (0)
        sort: Whether to sort the expression values (True)
        colorbar: Whether to show colorbar (None)
        col_cluster: If True, cluster the columns (False)
        row_cluster: If True, cluster the rows (False)
        context: Dictionary of parameters or preconfigured set name (None)
        font_scale: Scaling factor to scale the size of font elements (None)
        figsize: Figure size ((8, 4))
        show: Show the plot, do not return axis (None)
        save: If True or str, save the figure (None)
        kwargs: Arguments passed to seaborn's clustermap

    Returns:
        Matplotlib clustermap object

    """
    import seaborn as sns

    var_names = [name for name in var_names if name in adata.var_names]

    tkey, xkey = kwargs.pop("tkey", sortby), kwargs.pop("xkey", layer)
    time = adata.obs[tkey].values
    time = time[np.isfinite(time)]

    X = (
        adata[:, var_names].layers[xkey]
        if xkey in adata.layers.keys()
        else adata[:, var_names].X
    )
    if issparse(X):
        X = X.A
    df = pd.DataFrame(X[np.argsort(time)], columns=var_names)

    if n_convolve is not None:
        weights = np.ones(n_convolve) / n_convolve
        for gene in var_names:
            # TODO: Handle exception properly
            try:
                df[gene] = np.convolve(df[gene].values, weights, mode="same")
            except ValueError as e:
                print(f"Skipping variable {gene}: {e}")
                pass  # e.g. all-zero counts or nans cannot be convolved

    if sort:
        max_sort = np.argsort(np.argmax(df.values, axis=0))
        df = pd.DataFrame(df.values[:, max_sort], columns=df.columns[max_sort])
    strings_to_categoricals(adata)

    if col_color is not None:
        col_colors = to_list(col_color)
        col_color = []
        for _, col in enumerate(col_colors):
            if not is_categorical(adata, col):
                obs_col = adata.obs[col]
                cat_col = np.round(obs_col / np.max(obs_col), 2) * np.max(obs_col)
                adata.obs[f"{col}_categorical"] = pd.Categorical(cat_col)
                col += "_categorical"
                set_colors_for_categorical_obs(adata, col, palette)
            col_color.append(interpret_colorkey(adata, col)[np.argsort(time)])

    if "dendrogram_ratio" not in kwargs:
        kwargs["dendrogram_ratio"] = (
            0.1 if row_cluster else 0,
            0.2 if col_cluster else 0,
        )
    if "cbar_pos" not in kwargs or not colorbar:
        kwargs["cbar_pos"] = None

    kwargs.update(
        {
            "col_colors": col_color,
            "col_cluster": col_cluster,
            "row_cluster": row_cluster,
            "cmap": color_map,
            "xticklabels": False,
            "standard_scale": standard_scale,
            "figsize": figsize,
        }
    )

    args = {}
    if font_scale is not None:
        args = {"font_scale": font_scale}
        context = context or "notebook"

    with sns.plotting_context(context=context, **args):
        # TODO: Remove exception by requiring appropriate seaborn version
        try:
            cm = sns.clustermap(df.T, **kwargs)
        except ImportWarning:
            print("Please upgrade seaborn with `pip install -U seaborn`.")
            kwargs.pop("dendrogram_ratio")
            kwargs.pop("cbar_pos")
            cm = sns.clustermap(df.T, **kwargs)

    #savefig_or_show("heatmap", save=save, show=show)
    if show is False:
        return cm




# TODO: Add docstrings
def default_color(adata, add_outline=None):
    r"""Get default color for plotting based on available metadata.
    
    Arguments:
        adata: AnnData object
        add_outline: Outline parameter (None)
        
    Returns:
        Default color key string
    """
    if (
        isinstance(add_outline, str)
        and add_outline in adata.var.keys()
        and "recover_dynamics" in adata.uns.keys()
        and add_outline in adata.uns["recover_dynamics"]
    ):
        return adata.uns["recover_dynamics"][add_outline]
    return (
        "clusters"
        if "clusters" in adata.obs.keys()
        else "louvain"
        if "louvain" in adata.obs.keys()
        else "grey"
    )

# TODO: Add docstrings
def make_dense(X):
    r"""Convert sparse matrix to dense array.
    
    Arguments:
        X: Input matrix (sparse or dense)
        
    Returns:
        Dense numpy array
    """
    if issparse(X):
        XA = X.A if X.ndim == 2 else X.A1
    else:
        XA = X.A1 if isinstance(X, np.matrix) else X
    return np.array(XA)

# TODO: Add docstrings
def default_palette(palette=None):
    r"""Get default color palette for plotting.
    
    Arguments:
        palette: Input palette (None)
        
    Returns:
        Cycler object with color palette
    """
    from cycler import Cycler, cycler
    if palette is None:
        return rcParams["axes.prop_cycle"]
    elif not isinstance(palette, Cycler):
        return cycler(color=palette)
    else:
        return palette
    
# TODO: Add docstrings
def adjust_palette(palette, length):
    r"""Adjust palette to match required length.
    
    Arguments:
        palette: Input color palette
        length: Required number of colors
        
    Returns:
        Adjusted color palette
    """
    from cycler import Cycler, cycler
    islist = False
    if isinstance(palette, list):
        islist = True
    if (islist and len(palette) < length) or (
        not isinstance(palette, list) and len(palette.by_key()["color"]) < length
    ):
        if length <= 28:
            palette = sc.pl.palettes.zeileis_28
        elif length <= len(sc.pl.palettes.zeileis_102):  # 103 colors
            palette = sc.pl.palettes.zeileis_102
        else:
            palette = ["grey" for _ in range(length)]
            print("more than 103 colors would be required, initializing as 'grey'")
        return palette if islist else cycler(color=palette)
    elif islist:
        return palette
    elif not isinstance(palette, Cycler):
        return cycler(color=palette)
    else:
        return palette

# TODO: Add docstrings
def get_colors(adata, c):
    r"""Get colors for categorical or continuous variables.
    
    Arguments:
        adata: AnnData object
        c: Color specification
        
    Returns:
        Array of colors
    """
    if is_color_like(c):
        return c
    else:
        if f"{c}_colors" not in adata.uns.keys():
            palette = default_palette(None)
            palette = adjust_palette(palette, length=len(adata.obs[c].cat.categories))
            n_cats = len(adata.obs[c].cat.categories)
            adata.uns[f"{c}_colors"] = palette[:n_cats].by_key()["color"]
        if isinstance(adata.uns[f"{c}_colors"], dict):
            cluster_ix = adata.obs[c].values
        else:
            cluster_ix = adata.obs[c].cat.codes.values
        return np.array(
            [
                adata.uns[f"{c}_colors"][cluster_ix[i]]
                if cluster_ix[i] != -1
                else "lightgrey"
                for i in range(adata.n_obs)
            ]
        )

# TODO: Add docstrings
def interpret_colorkey(adata, c=None, layer=None, perc=None, use_raw=None):
    r"""Interpret color key specification for plotting.
    
    Arguments:
        adata: AnnData object
        c: Color key specification (None)
        layer: Layer to use for gene expression (None)
        perc: Percentile clipping values (None)
        use_raw: Whether to use raw data (None)
        
    Returns:
        Interpreted color values
    """
    if c is None:
        c = default_color(adata)
    if issparse(c):
        c = make_dense(c).flatten()
    if is_categorical(adata, c):
        c = get_colors(adata, c)
    elif isinstance(c, str):
        if is_color_like(c) and c not in adata.var_names:
            pass
        elif c in adata.obs.keys():  # color by observation key
            c = adata.obs[c]
        elif c in adata.var_names or (
            use_raw and adata.raw is not None and c in adata.raw.var_names
        ):  # by gene
            if layer in adata.layers.keys():
                if perc is None and any(
                    layer_name in layer
                    for layer_name in ["spliced", "unspliced", "Ms", "Mu", "velocity"]
                ):
                    perc = [1, 99]  # to ignore outliers in non-logarithmized layers
                c = adata.obs_vector(c, layer=layer)
            elif layer is not None and np.any(
                [
                    layer_name in layer or "X" in layer
                    for layer_name in adata.layers.keys()
                ]
            ):
                l_array = np.hstack(
                    [
                        adata.obs_vector(c, layer=layer)[:, None]
                        for layer in adata.layers.keys()
                    ]
                )
                l_array = pd.DataFrame(l_array, columns=adata.layers.keys())
                l_array.insert(0, "X", adata.obs_vector(c))
                c = np.array(l_array.astype(np.float32).eval(layer))
            else:
                if layer is not None and layer != "X":
                    print(layer, "not found. Using .X instead.")
                if adata.raw is None and use_raw:
                    raise ValueError("AnnData object does not have `raw` counts.")
                c = adata.raw.obs_vector(c) if use_raw else adata.obs_vector(c)
            c = c.A.flatten() if issparse(c) else c
        elif c in adata.var.keys():  # color by observation key
            c = adata.var[c]
        elif np.any([var_key in c for var_key in adata.var.keys()]):
            var_keys = [
                k for k in adata.var.keys() if not isinstance(adata.var[k][0], str)
            ]
            var = adata.var[list(var_keys)]
            c = var.astype(np.float32).eval(c)
        elif np.any([obs_key in c for obs_key in adata.obs.keys()]):
            obs_keys = [
                k for k in adata.obs.keys() if not isinstance(adata.obs[k][0], str)
            ]
            obs = adata.obs[list(obs_keys)]
            c = obs.astype(np.float32).eval(c)
        elif not is_color_like(c):
            raise ValueError(
                "color key is invalid! pass valid observation annotation or a gene name"
            )
        if not isinstance(c, str) and perc is not None:
            c = clip(c, perc=perc)
    else:
        c = np.array(c).flatten()
        if perc is not None:
            c = clip(c, perc=perc)
    return c

# TODO: Add docstrings
def clip(c, perc):
    r"""Clip values to specified percentile range.
    
    Arguments:
        c: Values to clip
        perc: Percentile range for clipping
        
    Returns:
        Clipped values
    """
    if np.size(perc) < 2:
        perc = [perc, 100] if perc < 50 else [0, perc]
    lb, ub = np.percentile(c, perc)
    return np.clip(c, lb, ub)

# TODO: Finish docstrings
def strings_to_categoricals(adata):
    r"""Transform string annotations to categoricals.
    
    Arguments:
        adata: AnnData object
        
    Returns:
        None (modifies adata in place)
    """
    from pandas import Categorical
    from pandas.api.types import is_bool_dtype, is_integer_dtype, is_string_dtype

    def is_valid_dtype(values):
        return (
            is_string_dtype(values) or is_integer_dtype(values) or is_bool_dtype(values)
        )

    df = adata.obs
    df_keys = [key for key in df.columns if is_valid_dtype(df[key])]
    for key in df_keys:
        c = df[key]
        c = Categorical(c)
        if 1 < len(c.categories) < min(len(c), 100):
            df[key] = c

    df = adata.var
    df_keys = [key for key in df.columns if is_string_dtype(df[key])]
    for key in df_keys:
        c = df[key].astype("U")
        c = Categorical(c)
        if 1 < len(c.categories) < min(len(c), 100):
            df[key] = c

# TODO: Add docstrings
def is_list(key):
    r"""Check if key is a list-like object.
    
    Arguments:
        key: Object to check
        
    Returns:
        Boolean indicating if key is list-like
    """
    return isinstance(key, (list, tuple, np.record))

# TODO: Add docstrings
def is_list_of_str(key, max_len=None):
    r"""Check if key is a list of strings.
    
    Arguments:
        key: Object to check
        max_len: Maximum length allowed (None)
        
    Returns:
        Boolean indicating if key is list of strings
    """
    if max_len is not None:
        return (
            is_list_or_array(key)
            and len(key) < max_len
            and all(isinstance(item, str) for item in key)
        )
    else:
        return is_list(key) and all(isinstance(item, str) for item in key)
    
# TODO: Add docstrings
def is_list_or_array(key):
    r"""Check if key is a list or array-like object.
    
    Arguments:
        key: Object to check
        
    Returns:
        Boolean indicating if key is list or array-like
    """
    return isinstance(key, (list, tuple, np.record, np.ndarray))


# TODO: Add docstrings
def to_list(key, max_len=20):
    r"""Convert key to list format.
    
    Arguments:
        key: Object to convert
        max_len: Maximum length allowed (20)
        
    Returns:
        List representation of key
    """
    from pandas import Index
    if isinstance(key, Index) or is_list_of_str(key, max_len):
        key = list(key)
    return key if is_list(key) and (max_len is None or len(key) < max_len) else [key]

# TODO: Add docstrings
def is_view(adata):
    r"""Check if AnnData object is a view.
    
    Arguments:
        adata: AnnData object
        
    Returns:
        Boolean indicating if adata is a view
    """
    return (
        adata.is_view
        if hasattr(adata, "is_view")
        else adata.isview
        if hasattr(adata, "isview")
        else adata._isview
        if hasattr(adata, "_isview")
        else True
    )

# TODO: Add docstrings
def is_categorical(data, c=None):
    r"""Check if data or column is categorical.
    
    Arguments:
        data: Data object to check
        c: Column name to check (None)
        
    Returns:
        Boolean indicating if data/column is categorical
    """
    from pandas.api.types import is_categorical_dtype as cat

    if c is None:
        return cat(data)  # if data is categorical/array
    if not is_view(data):  # if data is anndata view
        strings_to_categoricals(data)
    return isinstance(c, str) and c in data.obs.keys() and cat(data.obs[c])


# colors in addition to matplotlib's colors
additional_colors = {
    'gold2': '#eec900', 'firebrick3': '#cd2626', 'khaki2': '#eee685',
    'slategray3': '#9fb6cd', 'palegreen3': '#7ccd7c', 'tomato2': '#ee5c42',
    'grey80': '#cccccc', 'grey90': '#e5e5e5', 'wheat4': '#8b7e66', 'grey65': '#a6a6a6',
    'grey10': '#1a1a1a', 'grey20': '#333333', 'grey50': '#7f7f7f', 'grey30': '#4d4d4d',
    'grey40': '#666666', 'antiquewhite2': '#eedfcc', 'grey77': '#c4c4c4',
    'snow4': '#8b8989', 'chartreuse3': '#66cd00', 'yellow4': '#8b8b00',
    'darkolivegreen2': '#bcee68', 'olivedrab3': '#9acd32', 'azure3': '#c1cdcd',
    'violetred': '#d02090', 'mediumpurple3': '#8968cd', 'purple4': '#551a8b',
    'seagreen4': '#2e8b57', 'lightblue3': '#9ac0cd', 'orchid3': '#b452cd',
    'indianred 3': '#cd5555', 'grey60': '#999999', 'mediumorchid1': '#e066ff',
    'plum3': '#cd96cd', 'palevioletred3': '#cd6889'
}

# adapted from scanpy
def set_colors_for_categorical_obs(adata, value_to_plot, palette=None):
    r"""Set colors for categorical observation in AnnData object.

    Arguments:
        adata: AnnData object
        value_to_plot: Name of valid categorical observation
        palette: Color palette specification (None)
            Can be matplotlib colormap string, sequence of colors,
            or cycler object with 'color' key
            
    Returns:
        None (modifies adata.uns in place)
    """
    from matplotlib.colors import to_hex
    import matplotlib.pyplot as pl
    from cycler import Cycler, cycler


    color_key = f"{value_to_plot}_colors"
    valid = True
    categories = adata.obs[value_to_plot].cat.categories
    length = len(categories)

    if isinstance(palette, str) and "default" in palette:
        palette = sc.pl.palettes.zeileis_28 if length <= 28 else sc.pl.palettes.zeileis_102
    if isinstance(palette, str) and palette in adata.uns:
        palette = (
            [adata.uns[palette][c] for c in categories]
            if isinstance(adata.uns[palette], dict)
            else adata.uns[palette]
        )
    if palette is None and color_key in adata.uns:
        color_keys = adata.uns[color_key]
        # Check if colors already exist in adata.uns and if they are a valid palette
        if isinstance(color_keys, np.ndarray) and isinstance(color_keys[0], dict):
            adata.uns[color_key] = adata.uns[color_key][0]
        # Flatten the dict to a list (mainly for anndata compatibilities)
        if isinstance(adata.uns[color_key], dict):
            adata.uns[color_key] = [adata.uns[color_key][c] for c in categories]
        color_keys = adata.uns[color_key]
        for color in color_keys:
            if not is_color_like(color):
                # check if valid color translate to a hex color value
                if color in additional_colors:
                    color = additional_colors[color]
                else:
                    print(
                        f"The following color value found in "
                        f"adata.uns['{value_to_plot}_colors'] is not valid: '{color}'. "
                        f"Default colors will be used instead."
                    )
                    valid = False
                    break
        if len(adata.uns[color_key]) < len(adata.obs[value_to_plot].cat.categories):
            valid = False
    elif palette is not None:
        # check is palette given is a valid matplotlib colormap
        if isinstance(palette, str) and palette in pl.colormaps():
            # this creates a palette from a colormap. E.g. 'Accent, Dark2, tab20'
            cmap = pl.get_cmap(palette)
            colors_list = [to_hex(x) for x in cmap(np.linspace(0, 1, length))]

        else:
            # check if palette is an array of length n_obs
            if isinstance(palette, (list, np.ndarray)) or is_categorical(palette):
                if len(adata.obs[value_to_plot]) == len(palette):
                    cats = pd.Categorical(adata.obs[value_to_plot])
                    colors = pd.Categorical(palette)
                    if len(cats) == len(colors):
                        palette = dict(zip(cats, colors))
            # check if palette is as dict and convert it to an ordered list
            if isinstance(palette, dict):
                palette = [palette[c] for c in categories]
            # check if palette is a list and convert it to a cycler
            if isinstance(palette, abc.Sequence):
                if len(palette) < length:
                    print(
                        "Length of palette colors is smaller than the number of "
                        f"categories (palette length: {len(palette)}, "
                        f"categories length: {length}. "
                        "Some categories will have the same color."
                    )
                # check that colors are valid
                _color_list = []
                for color in palette:
                    if not is_color_like(color):
                        # check if valid color and translate to a hex color value
                        if color in additional_colors:
                            color = additional_colors[color]
                        else:
                            print(
                                f"The following color value is not valid: '{color}'. "
                                f"Default colors will be used instead."
                            )
                            valid = False
                            break
                    _color_list.append(color)
                palette = cycler(color=_color_list)

            if not isinstance(palette, Cycler) or "color" not in palette.keys:
                print(
                    "Please check that the value of 'palette' is a valid "
                    "matplotlib colormap string (eg. Set2), a list of color names or "
                    "a cycler with a 'color' key. Default colors will be used instead."
                )
                valid = False

            if valid:
                cc = palette()
                colors_list = [to_hex(next(cc)["color"]) for x in range(length)]
        if valid:
            adata.uns[f"{value_to_plot}_colors"] = colors_list
    else:
        # No valid palette exists or was given
        valid = False

    # Set to defaults:
    if not valid:
        # check if default matplotlib palette has enough colors
        if len(rcParams["axes.prop_cycle"].by_key()["color"]) >= length:
            cc = rcParams["axes.prop_cycle"]()
            palette = [next(cc)["color"] for _ in range(length)]
        # Else fall back to default palettes
        else:
            if length <= 28:
                palette = sc.pl.palettes.zeileis_28
            elif length <= len(sc.pl.palettes.zeileis_102):  # 103 colors
                palette = sc.pl.palettes.zeileis_102
            else:
                palette = ["grey" for _ in range(length)]
                print(
                    f"the obs value {value_to_plot!r} has more than 103 categories. "
                    f"Uniform 'grey' color will be used for all categories."
                )

        adata.uns[f"{value_to_plot}_colors"] = palette[:length]





import pandas as pd
import numpy as np
import anndata
import scanpy as sc
import torch
from typing import Union

from ..bulk import data_drop_duplicates_index,deseq2_normalize
from ..bulk2single import Bulk2Single
from ..single import scnocd
from ._vae import train_vae, generate_vae, load_vae

class BulkTrajBlend(object):
    r"""
    Bulk-to-single-cell trajectory blending using VAE and GNN integration.
    
    This class implements a comprehensive workflow for integrating bulk RNA-seq data
    with single-cell data to infer cellular trajectories and transition states.
    The method combines:
    - VAE-based bulk-to-single-cell deconvolution
    - Graph Neural Network (GNN) analysis for trajectory inference
    - Non-overlapping cell-type decomposition (NOCD) for transition states
    
    The workflow enables identification of intermediate cell states and trajectory
    dynamics that bridge bulk expression profiles with single-cell resolution.
    """

    def __init__(self,bulk_seq:pd.DataFrame,single_seq:anndata.AnnData,
                 celltype_key:str,bulk_group=None,max_single_cells:int=5000,
                 top_marker_num:int=500,ratio_num:int=1,gpu:Union[int,str]=0) -> None:
        r"""
        Initialize BulkTrajBlend for trajectory inference and cell blending.

        Arguments:
            bulk_seq: Bulk RNA-seq data with genes as rows and samples as columns
            single_seq: Single-cell RNA-seq reference data as AnnData object
            celltype_key: Column name in single_seq.obs containing cell type annotations
            bulk_group: Column names in bulk_seq for sample grouping (None)
            max_single_cells: Maximum number of single cells to use (5000)
            top_marker_num: Number of top marker genes per cell type (500)
            ratio_num: Cell selection ratio for each cell type (1)
            gpu: GPU device ID; -1 for CPU, 'mps' for Apple Silicon (0)
            
        Returns:
            None
        """

        self.bulk_seq = bulk_seq.copy()
        self.single_seq = single_seq.copy()
        self.celltype_key=celltype_key
        self.top_marker_num=top_marker_num
        self.ratio_num=ratio_num
        self.gpu=gpu
        self.group=bulk_group
        self.max_single_cells=max_single_cells
        if gpu=='mps' and torch.backends.mps.is_available():
            print('Note that mps may loss will be nan, used it when torch is supported')
            self.used_device = torch.device("mps")
        else:
            self.used_device = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')
        self.history=[]
        data_dg_v=self.bulk_seq.mean(axis=1)
        data_dg=pd.DataFrame(index=data_dg_v.index)
        data_dg['group']=data_dg_v
        self.bulk_seq_group=data_dg
        pass

    def bulk_preprocess_lazy(self,)->None:
        r"""
        Preprocess bulk RNA-seq data for trajectory analysis.
        
        Performs normalization, log transformation, and optional group averaging
        of bulk data for downstream trajectory inference.

        Arguments:
            None
            
        Returns:
            None: Updates self.bulk_seq and self.bulk_seq_group in place
        """

        print("......drop duplicates index in bulk data")
        self.bulk_seq=data_drop_duplicates_index(self.bulk_seq)
        print("......deseq2 normalize the bulk data")
        self.bulk_seq=deseq2_normalize(self.bulk_seq)
        print("......log10 the bulk data")
        self.bulk_seq=np.log10(self.bulk_seq+1)
        print("......calculate the mean of each group")
        if self.group is None:
            return None
        else:
            data_dg_v=self.bulk_seq[self.group].mean(axis=1)
            data_dg=pd.DataFrame(index=data_dg_v.index)
            data_dg['group']=data_dg_v
            self.bulk_seq_group=data_dg
        return None
    
    def single_preprocess_lazy(self,target_sum:int=1e4)->None:
        r"""
        Preprocess single-cell reference data for trajectory analysis.
        
        Normalizes single-cell data and makes cell/gene names unique for
        consistent integration with bulk data.

        Arguments:
            target_sum: Target sum for total count normalization (10000)
            
        Returns:
            None: Updates self.single_seq in place
        """

        print("......normalize the single data")
        self.single_seq.obs_names_make_unique()
        self.single_seq.var_names_make_unique()
        sc.pp.normalize_total(self.single_seq, target_sum=target_sum)
        print("......log1p the single data")
        sc.pp.log1p(self.single_seq)
        return None
    
    def vae_configure(self,cell_target_num=None,**kwargs):
        r"""
        Configure the VAE model for bulk-to-single-cell generation.
        
        Sets up the Bulk2Single model with cell-type target numbers either from
        deconvolution prediction or manual specification.

        Arguments:
            cell_target_num: Number of cells per type to generate (None for auto-prediction)
            **kwargs: Additional arguments passed to predicted_fraction method
            
        Returns:
            None: Initializes self.vae_model and preprocessing steps
        """
        self.vae_model=Bulk2Single(bulk_data=self.bulk_seq,single_data=self.single_seq,
                                   celltype_key=self.celltype_key,bulk_group=self.group,
                                      max_single_cells=self.max_single_cells,
                 top_marker_num=self.top_marker_num,ratio_num=self.ratio_num,gpu=self.gpu)
        if cell_target_num!=None:
            self.vae_model.cell_target_num=dict(zip(list(set(self.single_seq.obs[self.celltype_key])),
                                                [cell_target_num]*len(list(set(self.single_seq.obs[self.celltype_key])))))
        else:
            self.cellfract=self.vae_model.predicted_fraction(**kwargs)
        
        self.sc_ref=self.vae_model.sc_ref.copy()
        self.bulk_ref=self.vae_model.bulk_data.T.copy()

        self.vae_model.bulk_preprocess_lazy()
        self.vae_model.single_preprocess_lazy()
        self.vae_model.prepare_input()

        

    def vae_train(self,
                  vae_save_dir:str='save_model',
            vae_save_name:str='vae',
            generate_save_dir:str='output',
            generate_save_name:str='output',
            batch_size:int=512,
            learning_rate:int=1e-4,
            hidden_size:int=256,
            epoch_num:int=5000,
            patience:int=50,save:bool=True):
        r"""
        Train the VAE model for trajectory-aware single-cell generation.
        
        Trains the underlying Bulk2Single VAE model to generate synthetic single
        cells that preserve trajectory information from bulk data.

        Arguments:
            vae_save_dir: Directory to save trained VAE model ('save_model')
            vae_save_name: Filename for saved VAE model ('vae')
            generate_save_dir: Directory for generated data output ('output')
            generate_save_name: Filename for generated data ('output')
            batch_size: Training batch size (512)
            learning_rate: Optimizer learning rate (1e-4)
            hidden_size: Hidden layer dimensions (256)
            epoch_num: Maximum training epochs (5000)
            patience: Early stopping patience (50)
            save: Whether to save trained model (True)
            
        Returns:
            None: Updates self.vae_net with trained model
        """
        
        self.vae_net=self.vae_model.train(
                batch_size=batch_size,
                learning_rate=learning_rate,
                hidden_size=hidden_size,
                epoch_num=epoch_num,
                vae_save_dir=vae_save_dir,
                vae_save_name=vae_save_name,
                generate_save_dir=generate_save_dir,
                generate_save_name=generate_save_name,
                patience=patience,save=save)
        
        
    def vae_load(self,vae_load_dir:str,hidden_size:int=256):
        r"""
        Load a pre-trained VAE model for trajectory analysis.
        
        Loads a previously trained VAE model for generating trajectory-aware
        single-cell data.

        Arguments:
            vae_load_dir: Directory containing the trained VAE model
            hidden_size: Hidden layer dimensions matching training (256)
            
        Returns:
            None: Updates self.vae_net with loaded model
        """

        print(f'loading model from {vae_load_dir}')
        vae_net = self.vae_model.load(vae_load_dir,hidden_size=hidden_size)
        self.vae_net=vae_net

    def vae_generate(self,highly_variable_genes:bool=True,max_value:float=10,
                     n_comps:int=100,svd_solver:str='auto',leiden_size:int=50)->anndata.AnnData:
        r"""
        Generate trajectory-aware single-cell data with quality filtering.
        
        Uses the trained VAE to generate synthetic single cells and applies
        quality control filtering to remove noisy clusters.

        Arguments:
            highly_variable_genes: Whether to select highly variable genes (True)
            max_value: Maximum value for data scaling (10)
            n_comps: Number of principal components for PCA (100)
            svd_solver: SVD solver for PCA ('auto')
            leiden_size: Minimum cluster size threshold for filtering (50)

        Returns:
            anndata.AnnData: Generated and filtered single-cell data
        """

        generate_adata=self.vae_model.generate()
        self.generate_adata_raw=generate_adata.copy()
        generate_adata.raw = generate_adata
        if highly_variable_genes:
            sc.pp.highly_variable_genes(generate_adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
            generate_adata = generate_adata[:, generate_adata.var.highly_variable]
        sc.pp.scale(generate_adata, max_value=max_value)
        sc.tl.pca(generate_adata, n_comps=n_comps, svd_solver=svd_solver)
        sc.pp.neighbors(generate_adata, use_rep="X_pca")
        sc.tl.leiden(generate_adata)
        filter_leiden=list(generate_adata.obs['leiden'].value_counts()[generate_adata.obs['leiden'].value_counts()<leiden_size].index)
        generate_adata.uns['noisy_leiden']=filter_leiden
        print("The filter leiden is ",filter_leiden)
        generate_adata=generate_adata[~generate_adata.obs['leiden'].isin(filter_leiden)]
        self.generate_adata=generate_adata.copy()
        return generate_adata.raw.to_adata()
    
    def gnn_configure(self,use_rep='X',neighbor_rep='X_pca',
                      gpu=0,hidden_size:int=128,
                     weight_decay:int=1e-2,
                     dropout:float=0.5,
                     batch_norm:bool=True,
                     lr:int=1e-3,
                     max_epochs:int=500,
                     display_step:int=25,
                     balance_loss:bool=True,
                     stochastic_loss:bool=True,
                     batch_size:int=2000,num_workers:int=5,):
        r"""
        Configure Graph Neural Network for trajectory and transition state analysis.
        
        Sets up the NOCD (Non-Overlapping Cell-type Decomposition) GNN model
        for identifying cellular trajectories and intermediate states.

        Arguments:
            use_rep: Representation to use for GNN input ('X')
            neighbor_rep: Representation for neighbor graph construction ('X_pca')
            gpu: GPU device ID for training (0)
            hidden_size: Hidden layer dimensions in GNN (128)
            weight_decay: L2 regularization strength (1e-2)
            dropout: Dropout probability (0.5)
            batch_norm: Whether to use batch normalization (True)
            lr: Learning rate for GNN training (1e-3)
            max_epochs: Maximum training epochs (500)
            display_step: Frequency of progress updates (25)
            balance_loss: Whether to use balanced loss function (True)
            stochastic_loss: Whether to use stochastic loss (True)
            batch_size: Training batch size (2000)
            num_workers: Number of data loading workers (5)
            
        Returns:
            None: Initializes self.nocd_obj with configured GNN model
        """
        nocd_obj=scnocd(self.generate_adata,use_rep=use_rep,
                        neighbor_rep=neighbor_rep,gpu=gpu)
        #nocd_obj.device = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')
        nocd_obj.matrix_transform(clustertype=self.celltype_key)
        nocd_obj.matrix_normalize()
        nocd_obj.GNN_configure(hidden_size=hidden_size,weight_decay=weight_decay,
                             dropout=dropout,batch_norm=batch_norm,lr=lr,
                             max_epochs=max_epochs,display_step=display_step,
                             balance_loss=balance_loss,stochastic_loss=stochastic_loss,
                             batch_size=batch_size)
        nocd_obj.GNN_preprocess(num_workers=num_workers)

        self.nocd_obj=nocd_obj
    
    def gnn_train(self,thresh:float=0.5,gnn_save_dir:str='save_model',
            gnn_save_name:str='gnn'):
        r"""
        Train the GNN model for trajectory and transition state inference.
        
        Trains the NOCD model to identify cellular trajectories and overlapping
        cell communities representing transition states.

        Arguments:
            thresh: Threshold for community assignment (0.5)
            gnn_save_dir: Directory to save trained GNN model ('save_model')
            gnn_save_name: Filename for saved GNN model ('gnn')
            
        Returns:
            None: Trains model and computes trajectory results
        """
        self.nocd_obj.GNN_model()
        self.nocd_obj.GNN_result(thresh=thresh)
        self.nocd_obj.cal_nocd()
        self.nocd_obj.save(gnn_save_dir=gnn_save_dir,gnn_save_name=gnn_save_name)

    def gnn_load(self,gnn_load_dir:str,thresh:float=0.5,):
        r"""
        Load a pre-trained GNN model for trajectory analysis.
        
        Loads a previously trained NOCD model and computes trajectory results.

        Arguments:
            gnn_load_dir: Directory containing the trained GNN model
            thresh: Threshold for community assignment (0.5)
            
        Returns:
            None: Loads model and computes trajectory results
        """
        self.nocd_obj.load(gnn_load_dir)
        self.nocd_obj.GNN_result(thresh=thresh)
        self.nocd_obj.cal_nocd()
    
    def gnn_generate(self)->pd.DataFrame:
        r"""
        Generate overlapping cell communities representing transition states.
        
        Identifies and names cell communities based on trajectory analysis,
        creating binary matrix indicating community membership for each cell.

        Arguments:
            None
            
        Returns:
            pd.DataFrame: Binary matrix of cell community assignments with community names
        """
        '''
        pair_dict_r={}
        for i in range(self.nocd_obj.K):
            j=0
            while 1:
                if self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].shape[0]==0:
                    break
                if j>=len(self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index):
                    pair_dict_r[str(i)]=self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index[j-1]+'_'+str(j)
                    break
                if self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index[j] not in list(pair_dict_r.values()):
                    pair_dict_r[str(i)]=self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_n']==str(i)].obs.value_counts(self.celltype_key).index[j]
                    break
                else:
                    j+=1
        pair_dict_r
        '''
        unique_adata=self.nocd_obj.adata[~self.nocd_obj.adata.obs['nocd_n'].str.contains('-')]
        pair_dict_r={}
        repeat_celltype=dict(zip(list(set(unique_adata.obs[self.celltype_key])),np.zeros(len(list(set(unique_adata.obs[self.celltype_key]))))))
        for nocd_class in list(set(unique_adata.obs['nocd_n'])):
            now_celltype=unique_adata[unique_adata.obs['nocd_n']==nocd_class].obs.value_counts(self.celltype_key).index[0]
            if (now_celltype in pair_dict_r.values()):
                #print(now_celltype)
                pair_dict_r[str(nocd_class)]=now_celltype+'_'+str(int(repeat_celltype[now_celltype]))
                repeat_celltype[now_celltype]+=1
            else:
                pair_dict_r[str(nocd_class)]=now_celltype
                repeat_celltype[now_celltype]+=1

        def li_range(li,max_len):
            r=[0]*max_len   
            for i in li:
                r[int(i)]=1
            return r
        
        res_li=[li_range(i.split('-'),self.nocd_obj.K) for i in self.nocd_obj.adata.obs['nocd_n']]
        res_pd=pd.DataFrame(res_li,index=self.nocd_obj.adata.obs.index,columns=['nocd_'+i for i in [pair_dict_r[str(j)] for j in range(self.nocd_obj.K)]])
        print("The nocd result is ",res_pd.sum(axis=0))
        print("The nocd result has been added to adata.obs['nocd_']")
        self.nocd_obj.adata.obs=pd.concat([self.nocd_obj.adata.obs,res_pd],axis=1)
        return res_pd 
    
    def interpolation(self,celltype:str,adata:anndata.AnnData=None,)->anndata.AnnData:
        r"""
        Interpolate trajectory communities back to original data space.
        
        Integrates identified cell communities from generated data back with
        original single-cell reference data for downstream analysis.

        Arguments:
            celltype: Cell type or community name to interpolate
            adata: Original data for interpolation; uses self.single_seq if None (None)

        Returns:
            anndata.AnnData: Combined data with interpolated community information
        """
        if adata is None:
            adata=self.single_seq
        test_adata=self.nocd_obj.adata[self.nocd_obj.adata.obs['nocd_{}'.format(celltype)]==1].raw.to_adata()
        if test_adata.shape[0]!=0:
            adata1=anndata.concat([test_adata,
                        adata],merge='same')
        else:
            adata1=adata 
            print("The cell type {} is not in the nocd result".format(celltype))
        return adata1
    





import numpy as np
from sklearn.preprocessing import normalize
import matplotlib.pyplot as plt
from scipy.stats import linregress
from matplotlib.collections import LineCollection


from .binning_and_plotting import bin_data

from matplotlib.colors import ListedColormap
import seaborn as sns

# domain_boundary_label='domain boundaries \n from expression or \n cell type proportion'
domain_boundary_label='Domain boundaries'

# example of ct_colors: {'Oligodendrocytes': 'C6', 'Granule': 'mediumseagreen', 'Purkinje': 'red', 'Bergmann': 'C4', 'MLI1': 'gold', 'MLI2': 'goldenrod',  'Astrocytes': 'C0', 'Golgi': 'C9', 'Fibroblast': 'C5'}

# example of ct_pseudocounts: {3: 1} -- ie domain 4 needs CT pseudocount of 1

############################################################################################################

def domain_cts_svg(cell_type_df, gaston_labels, gaston_isodepth, domain_ct_threshold=0.7,
                  num_bins=60, num_bins_per_domain=[10,16,7,17]):
    N=len(gaston_labels)
    # need dummy counts mat and dummy gene_labels to create binning_output
    binning_output=bin_data(np.ones((N,10)), gaston_labels, gaston_isodepth, 
                         cell_type_df, np.array(['test' for i in range(10)]), num_bins=num_bins, num_bins_per_domain=num_bins_per_domain)
    ct_dict=get_domain_cts(binning_output, domain_ct_threshold) # {0: [list of CTs], 1: [list of CTs], ...}
    ct_list=[]
    for d in ct_dict:
        ct_list+=list(ct_dict[d])
    return np.unique(ct_list)

def get_domain_cts(binning_output, domain_ct_threshold, exclude_ct=[]):
    domain_ct_markers={}
    gaston_labels=binning_output['gaston_labels'] # unbinned labels
    cell_type_mat=binning_output['cell_type_mat'] # unbinned cell types
    cell_type_names=binning_output['cell_type_names']
    L=len(np.unique(gaston_labels))
    
    ct_names_not_excluded=np.array([ct for ct in cell_type_names if ct not in exclude_ct])
    ct_inds_not_excluded=np.array([i for i,ct in enumerate(cell_type_names) if ct not in exclude_ct])
    ct_mat_not_excluded=cell_type_mat[:, ct_inds_not_excluded]
    
    for t in range(L):
        pts_t=np.where(gaston_labels==t)[0]
        
        ct_counts_t=np.sum(ct_mat_not_excluded[pts_t,:],0)

        argsort_cts=np.argsort(ct_counts_t)[::-1]
        i=0
        while np.sum(ct_counts_t[argsort_cts[:i]]) / np.sum(ct_counts_t) < domain_ct_threshold:
            i+=1
        domain_t_cts=ct_names_not_excluded[argsort_cts[:i]]
        domain_ct_markers[t]=domain_t_cts
    return domain_ct_markers

############################################################################################################

# Make a plot of cell type proportion vs isodepth
# gaston_labels: N x 1 array of domain labels
# gaston_isodepth: N x 1 array of isodepth values per spot
# cell_type_df: N x C dataframe of (binary) cell type labels per spot
# (1)num_bins OR (2)num_bins_per_domain: 
#         specify either (1) number of total bins OR (2) number of bins in each spatial domain
# ct_list: list of cell types to plot
# ct_colors: list of colors for cell types in ct_list
# color_palette: matplotlib color palette, if ct_colors is not given
# ct_pseudocounts: dictionary of pseudocount to add to cell type proportion 
#         {3: 1} -- means domain 3 needs CT pseudocount of 1, so cell type c proportion is (c+1)/(n+1)
# domain_ct_threshold: we bold the cell types that cover at least domain_ct_threshold spots in each domain i
# ticksize: size of xticks, yticks
# domain_boundary_label: label for domain boundaries in legend
# exclude_ct: list of cell types to ignore
# width1, width2: width of cell types in their domain vs other domains
# lgd_width, lgd_fontsize: width/fontsize of legend
# include_lgd, lgd_frameon: boolean whether to include legend, and whether to include legend frame
# lgd_bbox: where to place legend
def plot_ct_props(cell_type_df, gaston_labels, gaston_isodepth, 
                  num_bins=60, num_bins_per_domain=[10,16,7,17],
                  ct_list=None, ct_colors=None, color_palette=None, 
                   ct_pseudocounts=None, linewidth=8, figsize=(15,6),
                   domain_ct_threshold=0.6, ticksize=25, domain_boundary_label=domain_boundary_label, 
                   exclude_ct=[],width1=8, width2=4,
                  lgd_width=4, include_lgd=True, lgd_fontsize=25, lgd_bbox=(1.75,1), lgd_frameon=True,
                 return_ct_raw=False):

    N=len(gaston_labels)
    # need dummy counts mat and dummy gene_labels to create binning_output
    binning_output=bin_data(np.ones((N,10)), gaston_labels, gaston_isodepth, 
                         cell_type_df, np.array(['test' for i in range(10)]), 
                         num_bins=num_bins, num_bins_per_domain=num_bins_per_domain)
    # print(binning_output)
    unique_binned_isodepths=binning_output['unique_binned_isodepths']
    binned_labels=binning_output['binned_labels']
    ct_count_mat=binning_output['binned_cell_type_mat'].T # len(unique_cell_types) x binned_labels
    cell_type_names=binning_output['cell_type_names']
    
    L=len(np.unique(binned_labels))
    
    left_bps=[]
    right_bps=[]

    for i in range(len(binned_labels)-1):
        if binned_labels[i] != binned_labels[i+1]:
            left_bps.append(unique_binned_isodepths[i])
            right_bps.append(unique_binned_isodepths[i+1])
            
    if ct_list is None:
        # ct_inds=np.array([i for i,ct in enumerate(cell_type_names) if ct not in exclude_ct])
        # ct_list=cell_type_names[ct_inds]
        domain_ct_markers=get_domain_cts(binning_output, domain_ct_threshold, exclude_ct=exclude_ct)
        ct_list=[]
        for v in domain_ct_markers.values():
            ct_list += list(v)
        ct_list=list(set(ct_list))
        ct_inds=np.array([np.where(cell_type_names==ct)[0][0] for ct in ct_list])
    else:
        ct_inds=np.array([np.where(cell_type_names==i)[0][0] for i in ct_list if i not in exclude_ct])
        ct_list=cell_type_names[ct_inds]

    if color_palette is None:
        l=len(ct_list)
        color_palette=ListedColormap([sns.color_palette("Spectral", as_cmap=True)(i/l) for i in range(l)])
    
    pc_mat=np.zeros( ct_count_mat.shape )
    if ct_pseudocounts is not None:
        for l in ct_pseudocounts:
            pc=ct_pseudocounts[l]
            pc_mat[:,np.where(binned_labels==l)[0]]=pc

    ct_count_mat=(ct_count_mat+pc_mat)[ct_inds]
    ct_count_prop=normalize(ct_count_mat,axis=0,norm='l1')

    fig,ax=plt.subplots(figsize=figsize)
        
    if ct_colors is None:
        ct_colors_list=[color_palette(i) for i in range(len(ct_list))]
    # print(len(ct_colors_list))
    c=0
    
    for i,ct in enumerate(ct_list):
        widths=np.ones(len(unique_binned_isodepths))*width2
        for l in range(L):
            if ct in domain_ct_markers[l]:
                pts_l=np.where(binned_labels==l)[0]
                widths[pts_l]=width1
        for s in range(len(widths)-1):
            if widths[s]==width1 and widths[s+1] < width1:
                widths[s]=width2
        
        x,y=unique_binned_isodepths, ct_count_prop[i,:]
        points = np.vstack((x, y)).T.reshape(-1, 1, 2)
        segments = np.hstack((points[:-1], points[1:]))
        if ct_colors is None:
            lc = LineCollection(segments, alpha=1, color=ct_colors_list[c], lw=widths, label=ct)
        else:
            lc = LineCollection(segments, alpha=1, color=ct_colors[ct], lw=widths, label=ct)
        line = ax.add_collection(lc)
        c+=1
        
    # AXES tick size
    ax.tick_params(axis='both', which='major', labelsize=ticksize)
    for i in range(L-1):
        if i==0:
            plt.axvline((left_bps[i]+right_bps[i])*0.5, color='black', ls='--', linewidth=3, label=domain_boundary_label)
        else:
            plt.axvline((left_bps[i]+right_bps[i])*0.5, color='black', ls='--', linewidth=3)

    if include_lgd:
        lgd=plt.legend(fontsize=lgd_fontsize, bbox_to_anchor=lgd_bbox, labelcolor='linecolor', frameon=lgd_frameon)
        try:
            for lh in lgd.legendHandles: 
                lh.set_alpha(1)
                lh.set_linewidth(lgd_width)
        except:
            for lh in lgd.legend_handles:
                lh.set_alpha(1)
                lh.set_linewidth(lgd_width)
        for text in lgd.get_texts():
            text.set_alpha(1)
            text.set_size(lgd_fontsize)
    
    
    plt.ylim((0, 1.05))
    sns.despine()

    if return_ct_raw:
        return ct_list, ct_count_prop, unique_binned_isodepths



import argparse
import os
import time

import numpy as np
import torch
from VAE_model import VAE
import sys
sys.path.append("..")
# from guided_diffusion.cell_datasets import load_data
# from guided_diffusion.cell_datasets_sapiens import load_data
# from guided_diffusion.cell_datasets_WOT import load_data
# from guided_diffusion.cell_datasets_muris import load_data
from ..guided_diffusion.cell_datasets_loader import load_data

torch.autograd.set_detect_anomaly(True)
import random

def seed_everything(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True


def prepare_vae(args, state_dict=None):
    """
    Instantiates autoencoder and dataset to run an experiment.
    """

    device = "cuda" if torch.cuda.is_available() else "cpu"

    datasets = load_data(
        data_dir=args["data_dir"],
        batch_size=args["batch_size"],
        train_vae=True,
    )

    autoencoder = VAE(
        num_genes=args["num_genes"],
        device=device,
        seed=args["seed"],
        loss_ae=args["loss_ae"],
        hidden_dim=128,
        decoder_activation=args["decoder_activation"],
    )
    if state_dict is not None:
        print('loading pretrained model from: \n',state_dict)
        use_gpu = device == "cuda"
        autoencoder.encoder.load_state(state_dict["encoder"], use_gpu)
        autoencoder.decoder.load_state(state_dict["decoder"], use_gpu)

    return autoencoder, datasets


def train_vae(args, return_model=False):
    """
    Trains a autoencoder
    """
    if args["state_dict"] is not None:
        filenames = {}
        checkpoint_path = {
            "encoder": os.path.join(
                args["state_dict"], filenames.get("model", "encoder.ckpt")
            ),
            "decoder": os.path.join(
                args["state_dict"], filenames.get("model", "decoder.ckpt")
            ),
            "gene_order": os.path.join(
                args["state_dict"], filenames.get("gene_order", "gene_order.tsv")
            ),
        }
        autoencoder, datasets = prepare_vae(args, checkpoint_path)
    else:
        autoencoder, datasets = prepare_vae(args)
   
    args["hparams"] = autoencoder.hparams

    start_time = time.time()
    for step in range(args["max_steps"]):

        genes, _ = next(datasets)

        minibatch_training_stats = autoencoder.train(genes)

        if step % 1000 == 0:
            for key, val in minibatch_training_stats.items():
                print('step ', step, 'loss ', val)

        ellapsed_minutes = (time.time() - start_time) / 60

        stop = ellapsed_minutes > args["max_minutes"] or (
            step == args["max_steps"] - 1
        )

        if ((step % args["checkpoint_freq"]) == 0 or stop):

            os.makedirs(args["save_dir"],exist_ok=True)
            torch.save(
                autoencoder.state_dict(),
                os.path.join(
                    args["save_dir"],
                    "model_seed={}_step={}.pt".format(args["seed"], step),
                ),
            )

            if stop:
                break

    if return_model:
        return autoencoder, datasets


def parse_arguments():
    """
    Read arguments if this script is called from a terminal.
    """

    parser = argparse.ArgumentParser(description="Finetune Scimilarity")
    # dataset arguments
    parser.add_argument("--data_dir", type=str, default='/data1/lep/Workspace/guided-diffusion/data/tabula_muris/all.h5ad')
    parser.add_argument("--loss_ae", type=str, default="mse")
    parser.add_argument("--decoder_activation", type=str, default="ReLU")

    # AE arguments                                             
    parser.add_argument("--local_rank", type=int, default=0)  
    parser.add_argument("--split_seed", type=int, default=1234)
    parser.add_argument("--num_genes", type=int, default=18996)
    parser.add_argument("--seed", type=int, default=0)
    parser.add_argument("--hparams", type=str, default="")

    # training arguments
    parser.add_argument("--max_steps", type=int, default=200000)
    parser.add_argument("--max_minutes", type=int, default=3000)
    parser.add_argument("--checkpoint_freq", type=int, default=50000)
    parser.add_argument("--batch_size", type=int, default=128)
    parser.add_argument("--state_dict", type=str, default="/data1/lep/Workspace/guided-diffusion/scimilarity-main/models/annotation_model_v1")  # if pretrain
    # parser.add_argument("--state_dict", type=str, default=None)   # if not pretrain

    parser.add_argument("--save_dir", type=str, default='../output/ae_checkpoint/muris_AE')
    parser.add_argument("--sweep_seeds", type=int, default=200)
    return dict(vars(parser.parse_args()))




import numpy as np
import anndata as ad
import pandas as pd


class SCENIC:

    def __init__(
            self,
            adata,
            db_glob,
            motif_path,
            n_jobs=8):
        
        from ctxcore.rnkdb import FeatherRankingDatabase as RankingDatabase
        self.adata = adata
        self.db_glob = db_glob
        self.motif_path = motif_path
        self.n_jobs = n_jobs

        import glob
        import os

        db_fnames = glob.glob(db_glob)  # e.g. your feather files
        dbs = [ RankingDatabase(fname, name=os.path.splitext(os.path.basename(fname))[0])
        for fname in db_fnames ]

        self.dbs = dbs


    def cal_grn(self,layer='counts'):
        import regdiffusion as rd
        if layer in self.adata.layers:
            x = self.adata.layers[layer].toarray()
            x = np.log(x+1.0)
        else:
            from ..pp import recover_counts
            
            if self.adata.X.max()<np.log1p(1e4):
                X_counts_recovered, size_factors_sub=recover_counts(self.adata.X, 1e4, 1e5, log_base=None, 
                                                          chunk_size=10000)
                self.adata.layers['counts']=X_counts_recovered
            elif self.adata.X.max()<np.log1p(50*1e4):
                X_counts_recovered, size_factors_sub=recover_counts(self.adata.X, 50*1e4, 50*1e5, log_base=None, 
                                                          chunk_size=10000)
                self.adata.layers['counts']=X_counts_recovered
            else:
                print('Please provide a layer with raw counts data')
                return None
        

            x = self.adata.layers['counts'].toarray()
            x = np.log(x+1.0)


        rd_trainer = rd.RegDiffusionTrainer(x)
        rd_trainer.train()
        grn = rd_trainer.get_grn(self.adata.var_names, top_gene_percentile=50)

        # Here for each gene, we are going to extract all edges
        edgelist = grn.extract_edgelist(k=-1, workers=self.n_jobs)
        edgelist.columns = ['TF', 'target', 'importance']
        self.edgelist = edgelist
        
        self.edgelist['importance']=self.edgelist['importance'].astype(np.float32)
        self.adjacencies = self.edgelist
        
        return edgelist
    
    def cal_regulons(
            self,
            rho_mask_dropouts=True,
            seed=42,**kwargs
        ):
        
        from ..externel.pyscenic.utils import modules_from_adjacencies
        from ..externel.pyscenic.prune import prune2df, df2regulons

        expr_mtx=self.adata.to_df()

        modules = list(
            modules_from_adjacencies(
                self.edgelist, 
                expr_mtx,
                rho_mask_dropouts=rho_mask_dropouts,
                **kwargs
            )
        )
        self.modules = modules

        # Calculate enriched motifs and build regulons
        df = prune2df(self.dbs, modules, self.motif_path,num_workers=self.n_jobs,
                        client_or_address='custom_multiprocessing')
        regulons = df2regulons(df)
        self.regulons = regulons

        from ..single._aucell import aucell

        auc_mtx = aucell(
            expr_mtx,
            regulons,
            num_workers=self.n_jobs,   # parallelism
            seed=seed                    # for reproducibility
        )
        self.auc_mtx = auc_mtx

        import anndata as ad

        ad_auc_mtx = ad.AnnData(auc_mtx)
        ad_auc_mtx.obs=self.adata.obs.loc[ad_auc_mtx.obs.index]
        self.ad_auc_mtx=ad_auc_mtx

        return ad_auc_mtx

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats
from sklearn.preprocessing import StandardScaler

def calculate_gene_temporal_center(rna_data, cell_indices, gene_name, 
                                 pseudotime_col='palantir_pseudotime', 
                                 method='weighted_mean'):
    """
    Calculate the temporal center position of a gene
    
    Parameters:
    -----------
    rna_data : AnnData object or similar structure
        Data object containing gene expression and cell metadata
    cell_indices : array-like
        Cell indices to analyze
    gene_name : str
        Gene name
    pseudotime_col : str
        Pseudotime column name
    method : str
        Calculation method: 'weighted_mean', 'peak', 'median'
    
    Returns:
    --------
    dict : Contains temporal center position and related statistical information
    """
    
    # Get gene expression data
    gene_expression = rna_data[cell_indices, gene_name].to_df()
    if isinstance(gene_expression, pd.DataFrame):
        gene_expression = gene_expression.iloc[:, 0]  # If DataFrame, take first column
    
    # Get corresponding pseudotime
    pseudotime = rna_data.obs.loc[cell_indices, pseudotime_col]
    
    # Ensure data alignment
    common_indices = gene_expression.index.intersection(pseudotime.index)
    gene_expr = gene_expression.loc[common_indices]
    pseudo_time = pseudotime.loc[common_indices]
    
    # Remove missing values
    valid_mask = ~(np.isnan(gene_expr) | np.isnan(pseudo_time))
    gene_expr = gene_expr[valid_mask]
    pseudo_time = pseudo_time[valid_mask]
    
    results = {}
    
    if method == 'weighted_mean':
        # Method 1: Expression level weighted average pseudotime
        weights = gene_expr + 1e-6  # Avoid division by zero
        temporal_center = np.average(pseudo_time, weights=weights)
        
    elif method == 'peak':
        # Method 2: Find pseudotime corresponding to expression peak
        # Use sliding window smoothing
        df = pd.DataFrame({'pseudotime': pseudo_time, 'expression': gene_expr})
        df_sorted = df.sort_values('pseudotime')
        
        # Calculate moving average
        window_size = max(10, len(df_sorted) // 20)
        df_sorted['expr_smooth'] = df_sorted['expression'].rolling(
            window=window_size, center=True).mean()
        
        # Find peak
        peak_idx = df_sorted['expr_smooth'].idxmax()
        temporal_center = df_sorted.loc[peak_idx, 'pseudotime']
        
    elif method == 'median':
        # Method 3: Pseudotime corresponding to expression median
        expr_threshold = np.median(gene_expr[gene_expr > 0])
        high_expr_cells = gene_expr >= expr_threshold
        temporal_center = np.median(pseudo_time[high_expr_cells])
    
    # Calculate additional statistical information
    results['temporal_center'] = temporal_center
    results['expression_range'] = (gene_expr.min(), gene_expr.max())
    results['pseudotime_range'] = (pseudo_time.min(), pseudo_time.max())
    results['correlation'] = stats.pearsonr(pseudo_time, gene_expr)[0]
    results['n_cells'] = len(gene_expr)
    
    # Calculate temporal distribution width of expression
    expr_weights = gene_expr / gene_expr.sum()
    variance = np.average((pseudo_time - temporal_center)**2, weights=expr_weights)
    results['temporal_width'] = np.sqrt(variance)
    
    return results

def plot_gene_temporal_pattern(rna_data, cell_indices, gene_name, 
                             pseudotime_col='palantir_pseudotime',
                             temporal_center=None):
    """
    Plot gene temporal expression pattern
    """
    # Get data
    gene_expression = rna_data[cell_indices, gene_name].to_df()
    if isinstance(gene_expression, pd.DataFrame):
        gene_expression = gene_expression.iloc[:, 0]
    
    pseudotime = rna_data.obs.loc[cell_indices, pseudotime_col]
    
    # Data alignment and cleaning
    common_indices = gene_expression.index.intersection(pseudotime.index)
    gene_expr = gene_expression.loc[common_indices]
    pseudo_time = pseudotime.loc[common_indices]
    
    valid_mask = ~(np.isnan(gene_expr) | np.isnan(pseudo_time))
    gene_expr = gene_expr[valid_mask]
    pseudo_time = pseudo_time[valid_mask]
    
    # Create plot
    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(10, 8))
    
    # Scatter plot
    ax1.scatter(pseudo_time, gene_expr, alpha=0.6, s=20)
    ax1.set_xlabel('Pseudotime')
    ax1.set_ylabel(f'{gene_name} Expression')
    ax1.set_title(f'{gene_name} Expression along Pseudotime')
    
    # Add temporal center line
    if temporal_center is not None:
        ax1.axvline(x=temporal_center, color='red', linestyle='--', 
                   label=f'Temporal Center: {temporal_center:.3f}')
        ax1.legend()
    
    # Smooth curve
    df = pd.DataFrame({'pseudotime': pseudo_time, 'expression': gene_expr})
    df_sorted = df.sort_values('pseudotime')
    window_size = max(10, len(df_sorted) // 20)
    df_sorted['expr_smooth'] = df_sorted['expression'].rolling(
        window=window_size, center=True).mean()
    
    ax1.plot(df_sorted['pseudotime'], df_sorted['expr_smooth'], 
            color='orange', linewidth=2, label='Smoothed')
    ax1.legend()
    
    # Expression distribution histogram
    ax2.hist(gene_expr, bins=30, alpha=0.7, edgecolor='black')
    ax2.set_xlabel(f'{gene_name} Expression')
    ax2.set_ylabel('Cell Count')
    ax2.set_title(f'{gene_name} Expression Distribution')
    
    plt.tight_layout()
    return fig

# Usage example
def analyze_gene_temporal_center(rna_margin, axial_cell_idx, gene_name='gata5'):
    """
    Complete gene temporal analysis workflow
    """
    print(f"Analyzing temporal center position of gene {gene_name}...")
    
    # Calculate temporal center using different methods
    methods = ['weighted_mean', 'peak', 'median']
    results = {}
    
    for method in methods:
        try:
            result = calculate_gene_temporal_center(
                rna_margin, axial_cell_idx, gene_name, method=method
            )
            results[method] = result
            print(f"\n{method} method:")
            print(f"  Temporal center position: {result['temporal_center']:.4f}")
            print(f"  Temporal distribution width: {result['temporal_width']:.4f}")
            print(f"  Expression-time correlation: {result['correlation']:.4f}")
            print(f"  Number of cells: {result['n_cells']}")
            
        except Exception as e:
            print(f"{method} method calculation failed: {e}")
    
    # Plot figure
    if 'weighted_mean' in results:
        fig = plot_gene_temporal_pattern(
            rna_margin, axial_cell_idx, gene_name,
            temporal_center=results['weighted_mean']['temporal_center']
        )
        plt.show()
    
    return results

def batch_calculate_gene_temporal_centers(rna_data, cell_indices, gene_list, 
                                        pseudotime_col='palantir_pseudotime',
                                        methods=['weighted_mean', 'peak', 'median'],
                                        min_expression=0.1, min_cells=10):
    """
    Batch calculate temporal center positions for multiple genes
    
    Parameters:
    -----------
    rna_data : AnnData object
        Data object containing gene expression and cell metadata
    cell_indices : array-like
        Cell indices to analyze
    gene_list : list
        List of genes to analyze
    pseudotime_col : str
        Pseudotime column name
    methods : list
        List of calculation methods
    min_expression : float
        Minimum expression threshold
    min_cells : int
        Minimum number of expressing cells
    
    Returns:
    --------
    pd.DataFrame : Table containing temporal center information for all genes
    """
    
    results_list = []
    failed_genes = []
    
    print(f"Starting batch calculation of temporal center positions for {len(gene_list)} genes...")
    
    for i, gene_name in enumerate(gene_list):
        if (i + 1) % 50 == 0:
            print(f"Progress: {i+1}/{len(gene_list)} genes")
        
        try:
            # Get gene expression data for pre-filtering
            gene_expression = rna_data[cell_indices, gene_name].to_df()
            if isinstance(gene_expression, pd.DataFrame):
                gene_expression = gene_expression.iloc[:, 0]
            
            # Check if gene meets filtering criteria
            expressing_cells = (gene_expression > min_expression).sum()
            max_expression = gene_expression.max()
            
            if expressing_cells < min_cells or max_expression < min_expression:
                print(f"Skipping gene {gene_name}: expressing cells={expressing_cells}, max expression={max_expression:.3f}")
                continue
            
            gene_result = {'gene': gene_name}
            
            # Calculate temporal center for each method
            for method in methods:
                try:
                    result = calculate_gene_temporal_center(
                        rna_data, cell_indices, gene_name, 
                        pseudotime_col=pseudotime_col, method=method
                    )
                    
                    # Add method-specific column names
                    gene_result[f'{method}_temporal_center'] = result['temporal_center']
                    gene_result[f'{method}_temporal_width'] = result['temporal_width']
                    
                    # Only add general information on first method (avoid duplication)
                    if method == methods[0]:
                        gene_result['expression_min'] = result['expression_range'][0]
                        gene_result['expression_max'] = result['expression_range'][1]
                        gene_result['pseudotime_min'] = result['pseudotime_range'][0]
                        gene_result['pseudotime_max'] = result['pseudotime_range'][1]
                        gene_result['expr_time_correlation'] = result['correlation']
                        gene_result['n_cells_analyzed'] = result['n_cells']
                        gene_result['expressing_cells'] = expressing_cells
                        
                except Exception as e:
                    print(f"Gene {gene_name} method {method} calculation failed: {e}")
                    gene_result[f'{method}_temporal_center'] = np.nan
                    gene_result[f'{method}_temporal_width'] = np.nan
            
            results_list.append(gene_result)
            
        except Exception as e:
            failed_genes.append(gene_name)
            print(f"Gene {gene_name} completely failed calculation: {e}")
    
    # Convert to DataFrame
    if results_list:
        df_results = pd.DataFrame(results_list)
        
        # Add some derived metrics
        if len(methods) > 1:
            # Calculate consistency between different methods
            temporal_cols = [f'{method}_temporal_center' for method in methods]
            df_results['temporal_center_std'] = df_results[temporal_cols].std(axis=1)
            df_results['temporal_center_range'] = (df_results[temporal_cols].max(axis=1) - 
                                                 df_results[temporal_cols].min(axis=1))
        
        # Calculate expression dynamic range
        df_results['expression_dynamic_range'] = (df_results['expression_max'] - 
                                                df_results['expression_min'])
        
        # Calculate expression density
        df_results['expression_density'] = (df_results['expressing_cells'] / 
                                          df_results['n_cells_analyzed'])
        
        # Sort by temporal center of first method
        main_method = methods[0]
        df_results = df_results.sort_values(f'{main_method}_temporal_center')
        df_results = df_results.reset_index(drop=True)
        
        print(f"\nBatch calculation completed!")
        print(f"Successfully analyzed: {len(df_results)} genes")
        print(f"Calculation failed: {len(failed_genes)} genes")
        
        return df_results, failed_genes
    
    else:
        print("No genes successfully calculated!")
        return pd.DataFrame(), failed_genes

def get_temporal_gene_clusters(df_results, method='weighted_mean', n_clusters=5):
    """
    Cluster genes based on temporal center positions
    """
    from sklearn.cluster import KMeans
    
    temporal_col = f'{method}_temporal_center'
    valid_data = df_results.dropna(subset=[temporal_col])
    
    if len(valid_data) < n_clusters:
        print(f"Valid genes ({len(valid_data)}) less than cluster number ({n_clusters})")
        return df_results
    
    # Clustering
    kmeans = KMeans(n_clusters=n_clusters, random_state=42)
    clusters = kmeans.fit_predict(valid_data[[temporal_col]])
    
    # Add clustering results
    df_results_copy = df_results.copy()
    df_results_copy['temporal_cluster'] = np.nan
    df_results_copy.loc[valid_data.index, 'temporal_cluster'] = clusters
    
    # Calculate statistics for each cluster
    cluster_stats = []
    for cluster_id in range(n_clusters):
        cluster_genes = df_results_copy[df_results_copy['temporal_cluster'] == cluster_id]
        if len(cluster_genes) > 0:
            stats = {
                'cluster': cluster_id,
                'n_genes': len(cluster_genes),
                'temporal_center_mean': cluster_genes[temporal_col].mean(),
                'temporal_center_std': cluster_genes[temporal_col].std(),
                'temporal_center_min': cluster_genes[temporal_col].min(),
                'temporal_center_max': cluster_genes[temporal_col].max()
            }
            cluster_stats.append(stats)
    
    cluster_stats_df = pd.DataFrame(cluster_stats)
    print("\nTemporal clustering statistics:")
    print(cluster_stats_df.to_string(index=False))
    
    return df_results_copy

def save_temporal_analysis_results(df_results, output_prefix='gene_temporal_analysis'):
    """
    Save analysis results to multiple files
    """
    # Save complete results
    df_results.to_csv(f'{output_prefix}_complete.csv', index=False)
    
    # Save simplified version (main metrics only)
    main_cols = ['gene', 'weighted_mean_temporal_center', 'weighted_mean_temporal_width',
                'expression_max', 'expr_time_correlation', 'expressing_cells', 'expression_density']
    available_cols = [col for col in main_cols if col in df_results.columns]
    df_summary = df_results[available_cols].copy()
    df_summary.to_csv(f'{output_prefix}_summary.csv', index=False)
    
    # Save early, middle, late gene lists
    if 'weighted_mean_temporal_center' in df_results.columns:
        temporal_col = 'weighted_mean_temporal_center'
        valid_results = df_results.dropna(subset=[temporal_col])
        
        # Divide into three parts by temporal center
        quantiles = valid_results[temporal_col].quantile([0.33, 0.67])
        
        early_genes = valid_results[valid_results[temporal_col] <= quantiles[0.33]]['gene'].tolist()
        middle_genes = valid_results[(valid_results[temporal_col] > quantiles[0.33]) & 
                                   (valid_results[temporal_col] <= quantiles[0.67])]['gene'].tolist()
        late_genes = valid_results[valid_results[temporal_col] > quantiles[0.67]]['gene'].tolist()
        
        # Save gene lists
        with open(f'{output_prefix}_early_genes.txt', 'w') as f:
            f.write('\n'.join(early_genes))
        with open(f'{output_prefix}_middle_genes.txt', 'w') as f:
            f.write('\n'.join(middle_genes))
        with open(f'{output_prefix}_late_genes.txt', 'w') as f:
            f.write('\n'.join(late_genes))
        
        print(f"\nTemporal gene classification:")
        print(f"Early genes ({len(early_genes)}): temporal center <= {quantiles[0.33]:.3f}")
        print(f"Middle genes ({len(middle_genes)}): {quantiles[0.33]:.3f} < temporal center <= {quantiles[0.67]:.3f}")
        print(f"Late genes ({len(late_genes)}): temporal center > {quantiles[0.67]:.3f}")
    
    print(f"\nResults saved to {output_prefix}_*.csv and {output_prefix}_*_genes.txt")

# Complete batch analysis workflow
def run_batch_temporal_analysis(rna_margin, axial_cell_idx, gene_list, 
                               output_prefix='gene_temporal_analysis'):
    """
    Run complete batch temporal analysis workflow
    
    Parameters:
    -----------
    rna_margin : AnnData object
        Data object
    axial_cell_idx : array-like  
        Cell indices
    gene_list : list
        Gene list
    output_prefix : str
        Output file prefix
    
    Returns:
    --------
    df_results : pd.DataFrame
        Analysis results table
    """
    
    # Batch calculation
    df_results, failed_genes = batch_calculate_gene_temporal_centers(
        rna_margin, axial_cell_idx, gene_list
    )
    
    if len(df_results) == 0:
        print("No genes successfully analyzed!")
        return None
    
    # Add clustering information
    df_results = get_temporal_gene_clusters(df_results)
    
    # Save results
    save_temporal_analysis_results(df_results, output_prefix)
    
    # Display results summary
    print(f"\n=== Analysis Results Summary ===")
    print(f"Total genes: {len(gene_list)}")
    print(f"Successfully analyzed: {len(df_results)}")
    print(f"Failed genes: {len(failed_genes)}")
    
    if len(df_results) > 0:
        print(f"\nTemporal center position statistics:")
        temporal_col = 'weighted_mean_temporal_center'
        if temporal_col in df_results.columns:
            print(f"Minimum: {df_results[temporal_col].min():.4f}")
            print(f"Maximum: {df_results[temporal_col].max():.4f}")
            print(f"Mean: {df_results[temporal_col].mean():.4f}")
            print(f"Median: {df_results[temporal_col].median():.4f}")
    
    return df_results

# Usage example:
"""
# Prepare gene list
gene_list = ['gata5', 'gata6', 'sox2', 'nanog', 'oct4']  # your gene list

# Run batch analysis
df_results = run_batch_temporal_analysis(
    rna_margin, 
    axial_cell_idx, 
    gene_list,
    output_prefix='axial_genes_temporal'
)

# View results
print(df_results.head())

# Filter genes in specific temporal range
early_genes_df = df_results[df_results['weighted_mean_temporal_center'] < 0.3]
print("Early genes:", early_genes_df['gene'].tolist())
"""

import numpy as np
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
import umap
from sklearn.neighbors import NearestNeighbors

def build_correlation_network_umap_layout(embedding_df, correlation_threshold=0.7, umap_neighbors=15, min_dist=0.1):
    """
    UMAP layout + correlation network
    
    Parameters:
    -----------
    embedding_df : pd.DataFrame
        Gene embedding data
    correlation_threshold : float
        Correlation threshold, control gene connection
    umap_neighbors : int
        UMAP neighbors parameter
    min_dist : float
        UMAP minimum distance parameter
    
    Returns:
    --------
    G : nx.Graph
        Network graph (based on correlation connection)
    pos : dict
        UMAP layout position
    correlation_matrix : pd.DataFrame
        Correlation matrix
    """
    
    # 1. Calculate correlation matrix
    from sklearn.metrics.pairwise import cosine_similarity
    correlation_matrix = cosine_similarity(embedding_df.values)
    correlation_df = pd.DataFrame(correlation_matrix, 
                                index=embedding_df.index, 
                                columns=embedding_df.index)
    
    # 2. UMAP dimensionality reduction to get layout position
    reducer = umap.UMAP(n_neighbors=umap_neighbors, min_dist=min_dist, random_state=42)
    umap_coords = reducer.fit_transform(embedding_df.values)
    
    # 3. Create network (based on correlation threshold)
    G = nx.Graph()
    genes = embedding_df.index.tolist()
    G.add_nodes_from(genes)
    
    # Add edges: based on correlation threshold
    for i, gene1 in enumerate(genes):
        for j, gene2 in enumerate(genes):
            if i < j:  # 避免重复
                correlation = float(correlation_df.loc[gene1, gene2])
                if abs(correlation) >= correlation_threshold:
                    G.add_edge(gene1, gene2, weight=abs(correlation))
    
    # 4. Create UMAP layout dictionary
    pos = {gene: umap_coords[i] for i, gene in enumerate(genes)}
    
    print(f"Network built successfully:")
    print(f"  Node number: {G.number_of_nodes()}")
    print(f"  Edge number: {G.number_of_edges()}")
    print(f"  Correlation threshold: {correlation_threshold}")
    
    return G, pos, correlation_df

def plot_umap_network(G, pos, node_colors=None, figsize=(12, 8)):
    """Simple plot of UMAP layout network"""
    
    plt.figure(figsize=figsize)
    
    # Plot edges
    nx.draw_networkx_edges(G, pos, alpha=0.3, width=0.5, edge_color='gray')
    
    # Plot nodes
    if node_colors is None:
        node_colors = 'lightblue'
    
    nx.draw_networkx_nodes(G, pos, node_color=node_colors, 
                          node_size=50, alpha=0.8)
    
    # Optional: add labels (if nodes are not many)
    if len(G.nodes()) < 50:
        nx.draw_networkx_labels(G, pos, font_size=8)
    
    plt.title('Gene Network with UMAP Layout')
    plt.axis('off')
    return plt.gcf()

def add_tf_regulation(G, tf_gene_dict):
    """
    Add TF-Gene regulation relationship to network
    
    Parameters:
    -----------
    G : nx.Graph
        Existing network
    tf_gene_dict : dict
        TF-Gene regulation dictionary {TF: [target_genes]}
    
    Returns:
    --------
    G : nx.Graph
        Network with regulation relationship
    tf_genes : list
        TF gene list
    """
    
    tf_genes = []
    regulation_edges = 0
    
    for tf, targets in tf_gene_dict.items():
        if tf in G.nodes():
            tf_genes.append(tf)
            
        for target in targets:
            if tf in G.nodes() and target in G.nodes():
                # Add regulation edge, marked as regulation type
                G.add_edge(tf, target, edge_type='regulation', weight=1.0)
                regulation_edges += 1
    
    print(f"Add regulation relationship:")
    print(f"  TF gene number: {len(tf_genes)}")
    print(f"  Regulation edge number: {regulation_edges}")
    
    return G, tf_genes

def plot_umap_network_with_regulation(G, pos, temporal_df, tf_genes, figsize=(15, 10)):
    """Plot network with regulation relationship"""
    
    plt.figure(figsize=figsize)
    
    # Separate different types of edges
    correlation_edges = [(u, v) for u, v, d in G.edges(data=True) 
                        if d.get('edge_type', 'correlation') == 'correlation']
    regulation_edges = [(u, v) for u, v, d in G.edges(data=True) 
                       if d.get('edge_type') == 'regulation']
    
    # Plot correlation edges (gray)
    nx.draw_networkx_edges(G, pos, edgelist=correlation_edges, 
                          alpha=0.3, width=0.5, edge_color='gray')
    
    # Plot regulation edges (red)
    nx.draw_networkx_edges(G, pos, edgelist=regulation_edges,
                          alpha=0.7, width=1.5, edge_color='red', 
                          style='dashed')
    
    # Get node colors (peak_temporal_center)
    node_colors = []
    for gene in G.nodes():
        if gene in temporal_df['gene'].values:
            temporal_val = float(temporal_df[temporal_df['gene'] == gene]['peak_temporal_center'].values[0])
            node_colors.append(temporal_val)
        else:
            node_colors.append(0)
    
    # Plot all nodes
    scatter = nx.draw_networkx_nodes(G, pos, node_color=node_colors, 
                                   node_size=80, alpha=0.8, 
                                   cmap='viridis', vmin=min(node_colors), vmax=max(node_colors))
    
    # Only show TF gene labels
    tf_pos = {gene: pos[gene] for gene in tf_genes if gene in pos}
    nx.draw_networkx_labels(G, tf_pos, font_size=8, font_color='black', 
                           font_weight='bold')
    
    # Add colorbar and legend
    plt.colorbar(scatter, label='Peak Temporal Center', shrink=0.8)
    
    # Add edge type legend
    from matplotlib.lines import Line2D
    legend_elements = [Line2D([0], [0], color='gray', lw=1, alpha=0.5, label='Correlation'),
                      Line2D([0], [0], color='red', lw=2, alpha=0.7, linestyle='--', label='TF Regulation')]
    plt.legend(handles=legend_elements, loc='upper right')
    
    plt.title('Gene Network: UMAP Layout + Correlation + TF Regulation')
    plt.axis('off')
    return plt.gcf()


def plot_grn(G, pos, tf_list,temporal_df, tf_gene_dict,
             figsize=(6,6),top_tf_target_num=5,title='GRN',ax=None,
             fontsize=12,cmap='RdBu_r',
             ):
    # Plot curved correlation edges
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)

    # Calculate PageRank
    pagerank = nx.pagerank(G)
        
    # Separate different types of edges
    correlation_edges = [(u, v) for u, v, d in G.edges(data=True) 
                        if d.get('edge_type', 'correlation') == 'correlation']
    regulation_edges = [(u, v) for u, v, d in G.edges(data=True) 
                    if d.get('edge_type') == 'regulation']

    from matplotlib.patches import FancyArrowPatch
    '''
    for edge in correlation_edges:
        x1, y1 = pos[edge[0]]
        x2, y2 = pos[edge[1]]
        
        arrow = FancyArrowPatch((x1, y1), (x2, y2),
                            connectionstyle="arc3,rad=0.1",  # 弯曲程度
                            arrowstyle="-", 
                            alpha=0.3, color='gray', linewidth=0.5)
        ax.add_patch(arrow)
    '''

    # Plot curved regulation edges
    for edge in regulation_edges:
        x1, y1 = pos[edge[0]]
        x2, y2 = pos[edge[1]]
        
        arrow = FancyArrowPatch((x1, y1), (x2, y2),
                            connectionstyle="arc3,rad=0.2",  # More curved
                            arrowstyle="-", 
                            alpha=0.7, color='#c2c2c2', linewidth=1.5,
                            #linestyle='dashed'
                            )
        ax.add_patch(arrow)

    # Get node colors
    node_colors = []
    for gene in G.nodes():
        if gene in temporal_df.index.values:
            temporal_val = float(temporal_df.loc[gene,'peak_temporal_center'])
            node_colors.append(temporal_val)
        else:
            node_colors.append(0)

    # Get node size (based on PageRank)
    node_sizes = []
    for gene in G.nodes():
        pr_value = pagerank[gene]
        # Scale PageRank value to appropriate node size range (50-500)
        size = 50 + (pr_value - min(pagerank.values())) / (max(pagerank.values()) - min(pagerank.values())) * 450
        node_sizes.append(size)

    # Plot nodes
    x_coords = [pos[node][0] for node in G.nodes()]
    y_coords = [pos[node][1] for node in G.nodes()]

    scatter = ax.scatter(x_coords, y_coords, c=node_colors, s=node_sizes, alpha=0.8, 
                        cmap=cmap, vmin=min(node_colors), vmax=max(node_colors))


    # Only show TF labels
    texts=[]
    from ..pl._palette import sc_color
    for tf,colors in zip(tf_list,sc_color[11:]):
        if tf in pos:
            # Add white border
            import matplotlib.patheffects as path_effects
            
            text = ax.text(pos[tf][0], pos[tf][1], tf, fontsize=10, 
            fontweight='normal', ha='center', va='center', color='white')
            text.set_path_effects([path_effects.Stroke(linewidth=3, foreground=colors),
                            path_effects.Normal()])
            texts.append(text)
            target_genes=tf_gene_dict[tf]
            for tf_g in target_genes[:top_tf_target_num]:
                if tf_g in pos:
                    text = ax.text(pos[tf_g][0], pos[tf_g][1], tf_g, fontsize=10, 
                    fontweight='normal', ha='center', va='center', color=colors)
                    text.set_path_effects([path_effects.Stroke(linewidth=3, foreground='white'),
                                    path_effects.Normal()])
                    texts.append(text)
            
            
            
    from adjustText import adjust_text
    adjust_text(texts,only_move={"text": "xy", "static": "xy", "explode": "xy", "pull": "xy"},
                        arrowprops=dict(arrowstyle='->', color='#a51616'))

    # Add colorbar and legend
    plt.colorbar(scatter, label='Mean\nExpression', shrink=0.5)

    #from matplotlib.lines import Line2D
    #legend_elements = [
    #    Line2D([0], [0], color='gray', lw=1, alpha=0.5, label='Correlation'),
    #    Line2D([0], [0], color='red', lw=2, alpha=0.7, linestyle='--', label='TF Regulation')
    #]
    #ax.legend(handles=legend_elements, loc='upper right')

    ax.set_title(title,fontsize=fontsize+1)
    ax.axis('off')
    return ax




"""Module for integrating spatial transcriptomics data across different conditions.

This module implements STAligner, a method for integrating spatial transcriptomics data
across different experimental conditions, technologies, and developmental stages. The
integration preserves both gene expression patterns and spatial organization.

Key features:
1. Spatial network construction
2. Graph neural network-based integration
3. Mutual nearest neighbor alignment
4. Batch effect correction
5. Cross-condition comparison

References:
    Zhou, X., Dong, K. & Zhang, S. Integrating spatial transcriptomics data 
    across different conditions, technologies and developmental stages. 
    Nat Comput Sci 3, 894–906 (2023)
"""
__author__ = "Xiang Zhou"
__email__ = "xzhou@amss.ac.cn"
__citation__ = "Zhou, X., Dong, K. & Zhang, S. Integrating spatial transcriptomics data across different conditions, technologies and developmental stages. Nat Comput Sci 3, 894–906 (2023)"

from ..externel.STAligner.mnn_utils import create_dictionary_mnn
from ..externel.STAligner.STALIGNER import STAligner

import numpy as np
import pandas as pd
from tqdm import tqdm
import scipy.sparse as sp
import sklearn.neighbors
import random

import torch
import torch.backends.cudnn as cudnn
cudnn.deterministic = True
cudnn.benchmark = True

import torch.nn.functional as F
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader

from .._settings import add_reference


def Cal_Spatial_Net(adata, rad_cutoff=None, k_cutoff=None,
                    max_neigh=50, model='Radius', verbose=True):
    r"""Construct spatial neighbor networks for spatial integration.
    
    This function builds a spatial neighborhood graph by connecting spots based
    on their physical distances. It supports both radius-based and k-nearest
    neighbor approaches for network construction.

    Arguments:
        adata: AnnData
            Input spatial data containing:
            - Spatial coordinates in obsm['spatial']
            - Gene expression data in X
        rad_cutoff: float, optional (default=None)
            Maximum distance between neighbors for 'Radius' model.
            Only used when model='Radius'.
        k_cutoff: int, optional (default=None)
            Number of nearest neighbors to connect for 'KNN' model.
            Only used when model='KNN'.
        max_neigh: int, optional (default=50)
            Maximum number of neighbors to consider during graph construction.
            Helps limit memory usage for large datasets.
        model: str, optional (default='Radius')
            Network construction method:
            - 'Radius': Connect spots within rad_cutoff distance
            - 'KNN': Connect k_cutoff nearest neighbors
        verbose: bool, optional (default=True)
            Whether to print network statistics.
        
    Returns:
        None
            Updates adata with:
            - adata.uns['Spatial_Net']: DataFrame of edges and distances
            - adata.uns['adj']: Sparse adjacency matrix

    Notes:
        - For STAligner, adjust rad_cutoff to ensure 5-10 neighbors per spot
        - Includes self-loops in adjacency matrix
        - Uses ball_tree algorithm for efficient neighbor search
        - Memory efficient implementation for large datasets
        - Critical for downstream integration tasks
    """
    assert (model in ['Radius', 'KNN'])
    if verbose:
        print('------Calculating spatial graph...')
    coor = pd.DataFrame(adata.obsm['spatial'])
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']

    nbrs = sklearn.neighbors.NearestNeighbors(
        n_neighbors=max_neigh + 1, algorithm='ball_tree').fit(coor)
    distances, indices = nbrs.kneighbors(coor)

    if model == 'KNN':
        indices = indices[:, 1:k_cutoff + 1]
        distances = distances[:, 1:k_cutoff + 1]
    if model == 'Radius':
        indices = indices[:, 1:]
        distances = distances[:, 1:]

    KNN_list = []
    for it in range(indices.shape[0]):
        KNN_list.append(pd.DataFrame(zip([it] * indices.shape[1], indices[it, :],distances[it, :])))
    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell1', 'Cell2', 'Distance']

    Spatial_Net = KNN_df.copy()
    if model == 'Radius':
        Spatial_Net = KNN_df.loc[KNN_df['Distance'] < rad_cutoff,]
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Net['Cell1'] = Spatial_Net['Cell1'].map(id_cell_trans)
    Spatial_Net['Cell2'] = Spatial_Net['Cell2'].map(id_cell_trans)

    if verbose:
        print(f'The graph contains {Spatial_Net.shape[0]} edges, {adata.n_obs} cells.')
        print(f'{(Spatial_Net.shape[0] / adata.n_obs):.4f} neighbors per cell on average.')
    adata.uns['Spatial_Net'] = Spatial_Net

    cells = np.array(adata.obs_names)
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    if 'Spatial_Net' not in adata.uns.keys():
        raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")
    Spatial_Net = adata.uns['Spatial_Net']
    G_df = Spatial_Net.copy()
    G_df['Cell1'] = G_df['Cell1'].map(cells_id_tran)
    G_df['Cell2'] = G_df['Cell2'].map(cells_id_tran)
    G = sp.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell1'], G_df['Cell2'])), shape=(adata.n_obs, adata.n_obs))
    G = G + sp.eye(G.shape[0])  # self-loop
    adata.uns['adj'] = G


class pySTAligner(object):
    r"""STAligner for spatial transcriptomics data integration.
    
    STAligner is a deep learning method for integrating spatial transcriptomics
    data across different experimental conditions, technologies, and developmental
    stages. It combines graph neural networks with mutual nearest neighbors to
    preserve both transcriptional and spatial relationships during integration.

    The method works by:
    1. Constructing spatial neighborhood graphs
    2. Learning batch-invariant embeddings
    3. Aligning similar regions across batches
    4. Preserving spatial organization
    5. Enabling cross-condition comparison

    Attributes:
        adata: AnnData
            Combined data containing all batches
        model: STAligner
            Neural network model for integration
        loader: DataLoader
            PyTorch geometric data loader
        device: torch.device
            Computing device (GPU/CPU)
        optimizer: torch.optim.Optimizer
            Adam optimizer for training

    Examples:
        >>> import scanpy as sc
        >>> import omicverse as ov
        >>> # Load data
        >>> adata1 = sc.read_visium(...)
        >>> adata2 = sc.read_visium(...)
        >>> # Construct spatial networks
        >>> ov.space.Cal_Spatial_Net(adata1, rad_cutoff=100)
        >>> ov.space.Cal_Spatial_Net(adata2, rad_cutoff=100)
        >>> # Combine data
        >>> adata = adata1.concatenate(adata2)
        >>> # Initialize STAligner
        >>> staligner = ov.space.pySTAligner(
        ...     adata=adata,
        ...     batch_key='batch',
        ...     Batch_list=[adata1, adata2]
        ... )
        >>> # Train model
        >>> staligner.train()
        >>> # Get integrated embeddings
        >>> embeddings = staligner.predicted()
    """
    
    def __init__(self,adata,
                 hidden_dims: list = [512, 30],
                 n_epochs: int = 1000,
                 lr: float = 0.001,
                 batch_key: str = 'batch_name',
                 key_added: str = 'STAligner',
                 gradient_clipping: float = 5,
                 weight_decay: float = 0.0001,
                 margin: float = 1,
                 verbose: bool = False,
                 random_seed: int = 666,
                 iter_comb = None,
                 knn_neigh: int = 100,
                 Batch_list = None,
                 device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
             ) -> None:
        r"""Initialize STAligner spatial integration model.
        
        This method sets up the STAligner model by:
        1. Processing input data
        2. Constructing graph neural networks
        3. Initializing optimization parameters
        4. Preparing batch alignment strategy

        Arguments:
            adata: AnnData
                Combined data containing all batches to integrate.
                Must have batch information in obs[batch_key].
            hidden_dims: list, optional (default=[512, 30])
                Dimensions of hidden layers in the neural network.
                The final dimension determines embedding size.
            n_epochs: int, optional (default=1000)
                Number of training epochs.
                More epochs may improve integration but take longer.
            lr: float, optional (default=0.001)
                Learning rate for Adam optimizer.
                Adjust if training is unstable.
            batch_key: str, optional (default='batch_name')
                Column in adata.obs containing batch information.
            key_added: str, optional (default='STAligner')
                Key for storing embeddings in adata.obsm.
            gradient_clipping: float, optional (default=5)
                Maximum gradient norm for stability.
            weight_decay: float, optional (default=0.0001)
                L2 regularization strength.
            margin: float, optional (default=1)
                Margin for triplet loss function.
                Larger values enforce stronger separation.
            verbose: bool, optional (default=False)
                Whether to print training progress.
            random_seed: int, optional (default=666)
                Random seed for reproducibility.
            iter_comb: list, optional (default=None)
                List of batch pairs to compare.
                If None, compares all pairs.
            knn_neigh: int, optional (default=100)
                Number of neighbors for mutual nearest neighbors.
            Batch_list: list, optional (default=None)
                List of individual batch AnnData objects.
                Must be in same order as in combined adata.
            device: torch.device, optional (default=auto)
                Computing device to use.
                Automatically uses GPU if available.

        Notes:
            - Requires pre-computed spatial networks
            - GPU acceleration recommended for large datasets
            - Batch_list order must match batch_key order
            - Memory usage scales with dataset size
            - Consider reducing knn_neigh for large datasets
        """
        self.device = device
        section_ids = np.array(adata.obs[batch_key].unique())

        comm_gene = adata.var_names
        data_list = []
        for adata_tmp in Batch_list:
            adata_tmp = adata_tmp[:, comm_gene]
            edge_index = np.nonzero(adata_tmp.uns['adj'])
            data_list.append(
                Data(edge_index=torch.LongTensor(np.array([edge_index[0], edge_index[1]])),
                              prune_edge_index=torch.LongTensor(np.array([])),
                              x=torch.FloatTensor(adata_tmp.X.todense())))

        loader = DataLoader(data_list, batch_size=1, shuffle=True)

        self.loader=loader
        self.adata = adata
        self.data_list = data_list

        # hyper-parameters
        self.lr=lr
        self.section_ids = section_ids
        self.n_epochs = n_epochs
        self.weight_decay=weight_decay
        self.hidden_dims = hidden_dims
        self.key_added = key_added
        self.gradient_clipping = gradient_clipping
        self.random_seed = random_seed
        self.margin = margin
        self.verbose = verbose
        self.iter_comb = iter_comb
        self.knn_neigh = knn_neigh
        self.Batch_list = Batch_list
        self.batch_key = batch_key
        self.model = STAligner(hidden_dims=[adata.X.shape[1], hidden_dims[0],
                                            hidden_dims[1]]).to(device)
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=lr,
                                          weight_decay=weight_decay)

        if verbose:
            print(self.model)

    def train(self):
        r"""Train the STAligner spatial integration model.
        
        This method performs two-stage training of the STAligner model:
        1. Pre-training with STAGATE to learn initial embeddings
        2. Fine-tuning with STAligner using triplet loss and MNN

        The training process:
        1. Sets random seeds for reproducibility
        2. Pre-trains with graph autoencoder
        3. Identifies mutual nearest neighbors
        4. Optimizes embeddings with triplet loss
        5. Monitors training progress
        6. Saves final embeddings

        Notes:
            - Progress shown if verbose=True
            - Uses GPU if available
            - Early stopping not implemented
            - Results stored in adata.obsm[key_added]
            - Memory usage increases during training
            - Consider batch size for large datasets
        """
        seed = self.random_seed
        random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)

        print('Pretrain with STAGATE...')
        for epoch in tqdm(range(0, 500)):
            for batch in self.loader:
                self.model.train()
                self.optimizer.zero_grad()
                batch = batch.to(self.device)
                z, out = self.model(batch.x, batch.edge_index)

                loss = F.mse_loss(batch.x, out)  # +adv_loss
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), 5.)
                self.optimizer.step()


        with torch.no_grad():
            z_list = []
            for batch in self.data_list:
                z, _ = self.model.cpu()(batch.x, batch.edge_index)
                z_list.append(z.cpu().detach().numpy())
        self.adata.obsm['STAGATE'] = np.concatenate(z_list, axis=0)
        self.model = self.model.to(self.device)


        print('Train with STAligner...')
        for epoch in tqdm(range(500, self.n_epochs)):
            if epoch % 100 == 0 or epoch == 500:
                if self.verbose:
                    print('Update spot triplets at epoch ' + str(epoch))
                with torch.no_grad():
                    z_list = []
                    for batch in self.data_list:
                        z, _ = self.model.cpu()(batch.x, batch.edge_index)
                        z_list.append(z.cpu().detach().numpy())

                self.adata.obsm['STAGATE'] = np.concatenate(z_list, axis=0)
                self.model = self.model.to(self.device)

                pair_data_list = []

                for comb in self.iter_comb:
                    #print(comb)
                    i, j = comb[0], comb[1]
                    batch_pair = self.adata[self.adata.obs[self.batch_key].isin([self.section_ids[i],
                                                                                  self.section_ids[j]])]
                    mnn_dict = create_dictionary_mnn(batch_pair, use_rep='STAGATE', batch_name=self.batch_key,
                                                           k=self.knn_neigh,
                                                           iter_comb=None, verbose=0)

                    batchname_list = batch_pair.obs[self.batch_key]
                    cellname_by_batch_dict = dict()
                    for batch_id in range(len(self.section_ids)):
                        cellname_by_batch_dict[self.section_ids[batch_id]] = batch_pair.obs_names[
                            batch_pair.obs[self.batch_key] == self.section_ids[batch_id]].values
                    anchor_list = []
                    positive_list = []
                    negative_list = []
                    for batch_pair_name in mnn_dict.keys():  # pairwise compare for multiple batches
                        for anchor in mnn_dict[batch_pair_name].keys():
                            anchor_list.append(anchor)
                            positive_spot = mnn_dict[batch_pair_name][anchor][0]
                            positive_list.append(positive_spot)
                            section_size = len(cellname_by_batch_dict[batchname_list[anchor]])
                            negative_list.append(
                                cellname_by_batch_dict[batchname_list[anchor]][np.random.randint(section_size)])

                    batch_as_dict = dict(zip(list(batch_pair.obs_names),
                                              range(0, batch_pair.shape[0])))
                    anchor_ind = list(map(lambda _: batch_as_dict[_], anchor_list))
                    positive_ind = list(map(lambda _: batch_as_dict[_], positive_list))
                    negative_ind = list(map(lambda _: batch_as_dict[_], negative_list))

                    edge_list_1 = np.nonzero(self.Batch_list[i].uns['adj'])

                    max_ind = edge_list_1[0].max()
                    edge_list_2 = np.nonzero(self.Batch_list[j].uns['adj'])

                    edge_list_2 = (edge_list_2[0] + max_ind + 1, edge_list_2[1] + max_ind + 1)
                    edge_list = [edge_list_1, edge_list_2]

                    edge_pairs = [np.append(edge_list[0][0], edge_list[1][0]),
                                   np.append(edge_list[0][1], edge_list[1][1])]

                    pair_data_list.append(Data(edge_index=torch.LongTensor(np.array([edge_pairs[0], edge_pairs[1]])),
                                           anchor_ind=torch.LongTensor(np.array(anchor_ind)),
                                           positive_ind=torch.LongTensor(np.array(positive_ind)),
                                           negative_ind=torch.LongTensor(np.array(negative_ind)),
                                           x=batch_pair.X)) 
                    #torch.FloatTensor(batch_pair.X.todense())

                # for temp in pair_data_list:
                #     temp.to(device)
                pair_loader = DataLoader(pair_data_list, batch_size=1, shuffle=True)

            for batch in pair_loader:
                self.model.train()
                self.optimizer.zero_grad()
                batch.x = torch.FloatTensor(batch.x[0].todense())
                batch = batch.to(self.device)
                z, out = self.model(batch.x, batch.edge_index)
                mse_loss = F.mse_loss(batch.x, out)

                anchor_arr = z[batch.anchor_ind,]
                positive_arr = z[batch.positive_ind,]
                negative_arr = z[batch.negative_ind,]

                triplet_loss = torch.nn.TripletMarginLoss(margin=self.margin, p=2, reduction='sum')
                tri_output = triplet_loss(anchor_arr, positive_arr, negative_arr)

                loss = mse_loss + tri_output
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.gradient_clipping)
                self.optimizer.step()

        add_reference(self.adata,'STAligner','spatial integration with STAligner')

    def predicted(self):
        r"""Generate and store the final embedding from trained STAligner model.
        
        Returns:
            AnnData object with STAligner embedding stored in obsm[key_added].
        """ 
        self.model.eval()
        with torch.no_grad():
            z_list = []
            for batch in self.data_list:
                z, _ = self.model.cpu()(batch.x, batch.edge_index)
                z_list.append(z.cpu().detach().numpy())

        self.adata.obsm[self.key_added] = np.concatenate(z_list, axis=0)
        add_reference(self.adata,'STAligner','spatial integration with STAligner')
        return self.adata
# End-of-file (EOF)



r"""
PopV algorithms collection for cell annotation and integration.

This module provides a comprehensive collection of algorithms for automated cell
type annotation and batch integration, implemented with a unified interface for
easy comparison and benchmarking.

Base classes:
    BaseAlgorithm: Abstract base class for all PopV algorithms

K-nearest neighbor methods:
    KNN_BBKNN: Batch-balanced k-nearest neighbors integration
    KNN_HARMONY: Harmony-based batch correction with kNN
    KNN_SCANORAMA: Scanorama integration with kNN classification
    KNN_SCVI: scVI latent space with kNN classification

Machine learning classifiers:
    CELLTYPIST: Automated cell type classification
    Random_Forest: Random forest classifier for cell types
    Support_Vector: Support vector machine classification
    XGboost: Gradient boosting classifier

Deep learning methods:
    SCANVI_POPV: Semi-supervised variational inference
    ONCLASS: Ontology-guided cell type annotation

Features:
- Unified interface for all algorithms
- Automatic hyperparameter optimization
- Cross-validation and benchmarking support
- Integration with popular single-cell frameworks
- Comprehensive evaluation metrics
"""

from ._base_algorithm import BaseAlgorithm
from ._bbknn import KNN_BBKNN
from ._celltypist import CELLTYPIST
from ._harmony import KNN_HARMONY
from ._onclass import ONCLASS
from ._rf import Random_Forest
from ._scanorama import KNN_SCANORAMA
from ._scanvi import SCANVI_POPV
from ._scvi import KNN_SCVI
from ._svm import Support_Vector
from ._xgboost import XGboost

__all__ = [
    "BaseAlgorithm",
    "CELLTYPIST",
    "KNN_BBKNN",
    "KNN_HARMONY",
    "KNN_SCANORAMA",
    "KNN_SCVI",
    "ONCLASS",
    "Random_Forest",
    "SCANVI_POPV",
    "Support_Vector",
    "XGboost",
]
